/***********************************************************************
 * Bdbj 1.2.1 (by Jon Cook August 2002)
 *
 * Bridge.java 
 * 
 * This class is just to make my life a bit easier, static
 * methods are easier (and quicker) to lookup;  but we don't
 * want this hack to uglify BdbjInterpreter or BdbjDebugThread.
 *
 ***********************************************************************/

package jjc.bdbj.client;

public class Bridge {
    
    /* The idea is this: we don't want to use the JNI to call the
     * C (cos of threading), so instead we use a class loader to run the
     * target.   The client and target occupy the same VM (they
     * always did), so the C can lookup client methods and call them.
     * Of course, now the Java cannot directly call the C.
     * There is only one occassion where we really need to get
     * info across with the Java side controlling the timing,
     * and that is to communicate the source files/init line/classname.
     * So we call an empty method in the client, which is detected by
     * the C when it is executed (in runVirtualMachine() ) and does
     * the necessary callback.   All other communication is done
     * by the C calling Java.  This was in fact the case when we
     * used the JNI with the exception of breakpoints, but that is
     * easily fixed (load table on receipt of each breakpoint command).
     */

    static Interpreter interpreter = null;
    static Breakpoints breakpoints = null;

    static void setInterpreter( Interpreter interpreter ) {
	Bridge.interpreter = interpreter;
    }

    static void setBreakpoints( Breakpoints breakpoints ) {
	Bridge.breakpoints = breakpoints;
    }

    static void triggerInitialValueRequest( ) {
	// this call is detected in the modified Kaffe, which then takes the hint and does the necessary. 
	;
    }

    static Object getBreakpointObject( ) {
	return (Object)breakpoints;
    }

    /* Callbacks to get initial values */

    public static String getInitialFile( ) {
	return interpreter.getInitialFile( );
    }

    public static String getDebugFiles( ) {
	return interpreter.getDebugFiles( );
    }

    public static int getInitialLine( ) {
	return interpreter.getInitialLine( );
    }

    public static int getBufferSize( ) {
	return interpreter.getBufferSize( );
    }

    public static String getDeepList( ) {
	return interpreter.getDeepLogList( ).replace( '.', '/' );
    }

    /* Callbacks to load the breakpoint table on receiving
     * a breakpoint command */

    public static int getBreakpointLine( int index ) {
	return breakpoints.getBreakpointLine( index );
    }

    public static String getBreakpointFile( int index ) {
	return breakpoints.getBreakpointFile( index );
    }

    public static String getBreakpointString( ) {
	return breakpoints.getBreakpointString( );
    }
    
    /* Callbacks to report thread birth and death */

    public static Object newThread( String threadName, int thr, int shouldCreate ) {
	try {
	    interpreter.newThread( threadName, thr, shouldCreate );
	    return interpreter.debugThreads[ thr ].objects;
	} catch ( NotEnoughMemoryException e ) {
	    return null;
	}
    }

    public static Object newMiniThread( int thr, int shouldCreate ) {
	try {
	    interpreter.newMiniThread( thr, shouldCreate );
	    return interpreter.debugThreads[ thr ].objects;
	} catch ( NotEnoughMemoryException e ) {
	    return null;
	}
    }

    public static void threadDead( int thr, int shouldDelete ) {
	interpreter.threadDead( thr, shouldDelete );
    }

    public static void miniThreadDead( int thr, int shouldDelete ) {
	interpreter.miniThreadDead( thr, shouldDelete );
    }

    /* Callbacks to report back */

    public static void reportNewWatches( String watch ) {
	interpreter.reportNewWatches( watch );
    }

    public static void reportWatchIntegral( String var, long val ) {
	interpreter.reportNewWatches( var + " == " + val );
    }

    public static String changeNewlines( String sin ) {
	StringBuffer sb = new StringBuffer( "" );
	for( int i = 0; i < sin.length( ); i++ ) {
	    char c = sin.charAt( i );
	    switch( c ) {
	    default:  sb.append( c ); break; 
		case '\b': sb.append( "\\b" ); break;
		case '\t': sb.append( "\\t" ); break;
		case '\n': sb.append( "\\n" ); break;
		case '\f': sb.append( "\\f" ); break;
		case '\r': sb.append( "\\r" ); break;
		case '\"': sb.append( "\\\"" ); break;
                case '\'': sb.append( "\\\'" ); break;
		case '\\': sb.append( "\\\\" ); break;
	    }
	}
	return sb.toString( );
    }

    public static void reportWatchObject( String var, Object obj ) {
	if( obj instanceof String )
	    interpreter.reportNewWatches( var + " == \"" + 
					  changeNewlines( obj.toString( ) ) + "\"" );
	else {
	    try {
		interpreter.reportNewWatches( var + " == {" + 
					      changeNewlines( obj.toString( ) ) + "}" );
	    } catch( Throwable t ) {
		interpreter.reportNewWatches( var + " == <error in toString()>" );
	    }
	}
    }

    public static void reportWatchFloat( String var, float fl ) {
	interpreter.reportNewWatches( var + " == " + fl );
    }

    public static void reportWatchDouble( String var, double d ) {
	interpreter.reportNewWatches( var + " == " + d );
    }

    public static void reportWatchChar( String var, char ch ) {
	interpreter.reportNewWatches( var + " == \'" + ch + "\'" );
    }

    public static void reportWatchBoolean( String var, boolean b ) {
	interpreter.reportNewWatches( var + " == " + b );
    }

    public static void reportNewLine( int line, int fileOffset, int thr, int gotoLine ) {
	interpreter.reportNewLine( line, fileOffset, thr, gotoLine );
    }

    public static void reportNewPc( int pc, String meth, String sig, int fileOffset, int thr, int gotoLine ) {
	interpreter.reportNewPc( pc, meth + sig, fileOffset, thr, gotoLine );
    }

    public static void writePrompt( int thr ) {
	interpreter.writePrompt( thr );
    }

    public static void cantBack( ) {
	interpreter.cantBack( );
    }

    /* Callbacks to request instructions */

    public static int getCommand( int thr ) {
	return interpreter.getCommand( thr );
    }

    public static String getWatches( int thr ) {
	return interpreter.getWatches( thr );
    }

    public static String getNextBreakpointString( int thr ) {
	return interpreter.getNextBreakpointString( thr );
    }

    public static void askFullThrow( Object exn, int thr ) {
	interpreter.askFullThrow( exn, thr );
    }

    /* toString call protected against exceptions
     * primary concern is calls to ToString before init resulting in null pointer exceptions 
     */

    public static String safeToString( Object o ) {
	try {
	    return "{\"" + o.toString( ) + "\"}";
	} catch( Throwable t ) {
	    return "<exception in toString()>";
	}
    }

    public static long freeMemory( ) {
	return Runtime.getRuntime( ).freeMemory( );
    }
}
