/***********************************************************************
 * Bdbj 1.2.1 (by Jon Cook August 2002)
 *
 * DebugThread.java 
 * 
 * Receive instructions from the UI for a particular thread and pass
 * them on to the corresponding thread in Kaffe.
 ***********************************************************************/

package jjc.bdbj.client;

public class DebugThread {

    final static int STEP          = 1;
    final static int BREAKPOINT    = 2;
    final static int STEPOVER      = 3;
    final static int STEPBACK      = 5;
    final static int STEPBACKOVER  = 6;
    final static int GETWATCH      = 8; // new
    final static int BACKBPT       = 9; // new

    final static int BYTECODE_STEP        = 10;
    final static int BYTECODE_STEPOVER    = 11;
    final static int BYTECODE_STEPBACK    = 12;
    final static int BYTECODE_STEPBACKOVER= 13;
    final static int BYTECODE_BREAKPOINT  = 14;
    final static int BYTECODE_BACKBPT     = 15;

    final static int GCBUFFER      = 16;

    public Object[] objects = null;

    private boolean active = false;
    private int threadNumber;
    private String threadName = "<no name>"; 

    private String fileList = null;
    private boolean prompt;

    private Interpreter interpreter = null;
    private ClientSocket clientSocket = null;

    private SharedFIFOQueue commandBuffer = new SharedFIFOQueue( );
    private SharedFIFOQueue watchBuffer = new SharedFIFOQueue( );
    private SharedFIFOQueue breakpointBuffer = new SharedFIFOQueue( );

    private static boolean lineReportedFlag = false;
    private static void setLineReportedFlag( boolean b ) {
	synchronized( DebugThread.class ) {
	    lineReportedFlag = b;
	}
    }
    private static boolean getLineReportedFlag( ) {
	synchronized( DebugThread.class ) {
	    return lineReportedFlag;
	}
    }

    public DebugThread( Interpreter interpreter, ClientSocket clientSocket, 
			    String fileList, boolean prompt, int threadNumber, int bufferSize ) {
	this.interpreter = interpreter;
	this.clientSocket = clientSocket;
	this.fileList = fileList;
	this.prompt = prompt;
	this.threadNumber = threadNumber;
	objects = new Object[ bufferSize ];
    }

    /*
    public void finalize( ) {
	System.out.println( "Thread " + threadNumber + " object array gc'd" );
    }
    */

    private String removeNewlines( String string ) {
	String output = "";
	for( int i = 0; i < string.length( ); i++ )
	    if( string.charAt( i ) == '\n' )
		output += "\\n";
	    else if( string.charAt( i ) == '\r' )
		output += "\\r";
	    else 
		output += string.charAt( i );
	return output;
    }

    // Called from C:

    public void newThread( String threadName ) {
	threadName = removeNewlines( threadName );
	this.threadName = threadName;
	clientSocket.writeLine( "new-thread " + threadNumber + " " + threadName );
	active = true;
    }

    public void threadDead( ) {
	clientSocket.writeLine( "thread-dead " + threadNumber + " " + threadName );
	commandBuffer.removeAll( );
	watchBuffer.removeAll( );
	breakpointBuffer.removeAll( );
	active = false;
	waitPrompt( );
    }


    private String getClassname( int fileOffset ) {
	String end = fileList.substring( fileOffset + 1 );
	int pos = end.indexOf( " " );
	return end.substring( 0, pos );
    }

    public void reportNewLine( int line, int fileOffset, int gotoLine ) {
	clientSocket.writeLine( ( gotoLine == 0 ? "" : "goto " ) + "line " + line + 
				" in-file " + getClassname( fileOffset ).trim( ) +
				" on-thread " + threadNumber + " " + threadName );
    }

    public void reportNewPc( int pc, String meth, int fileOffset, int gotoLine ) {
	clientSocket.writeLine( ( gotoLine == 0 ? "" : "goto " ) + "pc " + pc + 
				" in-method " + meth +
				" in-file " + getClassname( fileOffset ).trim( ) + 
 				" on-thread " + threadNumber + " " + threadName );
    }

    public void writePrompt( ) { 
	waitPrompt( );
    }

    public int getCommand( ) { 
	return ( (Integer)commandBuffer.get( ) ).intValue( );
    }

    public String getWatches( ) { // called from C
        return ( (String)watchBuffer.get( ) ).toString( );        
    }

    public String getBreakpointString( ) {
	return ( (String)breakpointBuffer.get( ) ).toString( );
    }

    // End of called from C.

    public void putCommand( int command ) {
        commandBuffer.put( new Integer( command ) );
    }

    public void putWatch( String watch ) {
        watchBuffer.put( watch );
    }

    public void putBreakpoints( String breakpoints ) {
	breakpointBuffer.put( breakpoints );
    }

    public void interpretCommand( String command ) {
	if( command.equals( "" ) ) 
	    ;
	else if( command.equals( "bcstep" ) ) {
	    putCommand( BYTECODE_STEP );
	}
	else if( command.equals( "bcback" ) ) {
	    putCommand( BYTECODE_STEPBACK );
	}
	else if( command.equals( "bcstepover" ) ) {
	    putCommand( BYTECODE_STEPOVER );
	} 
	else if( command.equals( "bcbackover" ) ) {
	    putCommand( BYTECODE_STEPBACKOVER );
	}
	else if( command.equals( "bcbpt" ) ) {
	    putBreakpoints( interpreter.breakpoints.getBreakpointString( ) );
	    putCommand( BYTECODE_BREAKPOINT );
	} 
	else if( command.equals( "bcbackbpt" ) ) {
	    putBreakpoints( interpreter.breakpoints.getBreakpointString( ) );
	    putCommand( BYTECODE_BACKBPT );
	}
        else if( command.equals( "step" ) ) { 
            putCommand( STEP );
        }
        else if( command.equals( "stepover" ) ) {
            putCommand( STEPOVER );
        }
        else if( command.equals( "back" ) ) {
            putCommand( STEPBACK );
        }
        else if( command.equals( "backover" ) ) {
            putCommand( STEPBACKOVER );
        }
        else if( command.equals( "bpt" ) ) {
	    putBreakpoints( interpreter.breakpoints.getBreakpointString( ) );
            putCommand( BREAKPOINT );
        }
	else if( command.equals( "backbpt" ) ) {
	    putBreakpoints( interpreter.breakpoints.getBreakpointString( ) );
	    putCommand( BACKBPT );
	}
	else if( command.equals( "gcbuffer" ) ) {
	    putCommand( GCBUFFER );
	}
        else if( command.startsWith( "watch" ) ) {
            String identifier = Decoder.getString( command );
            if( !Watches.isValidIdentifier( identifier ) ) 
		clientSocket.writeLine( "watch " + identifier + " == <invalid identifier>" );
            else {
                putWatch( identifier );
                putCommand( GETWATCH );
            }
        }
        else
            clientSocket.writeLine( "error: Unrecognised Command on thread " + threadNumber + " " + threadName );
    }

    /* We want to print the "Bdbj >" prompt at the right time, but...
     * We have the problem that we don't really know how much stuff is going
     * to come out after a step/stepover etc., as it might create loads of
     * new threads which all print there "new-thread" and first "line"
     * messages.   When the current thread prints its new line message
     * we reason that the other messages, if there are any, will come
     * very soon (but there might be lots of them).   So we wait for
     * 100ms and if no new messages have come out we print the prompt,
     * if they have, we wait for another 100ms and so on.   This gives
     * a slightly palpable delay when command line debugging (no effect
     * on the GUI), but almost always ensures that the prompt it printed
     * after all the output for the command.   Also note that new line
     * messages are not going to keep coming for ever (at most 15) so
     * there is no risk of this thread failing to terminate.
     * The thread is not daemonic because if the step,... command
     * terminates the program we want to print the prompt demanded by
     * the "thread-dead" message before BdbjClient terminates.
     */

    private void waitPrompt( ) {
	if( prompt ) {
	    if( interpreter.amICurrentThread( this ) ) {
		new Thread( ) {
			public void run( ) {
			    for( ; ; ) {
				setLineReportedFlag( false );
				try {
				    sleep( 100 );
				} catch( InterruptedException _ ) {
				}
				if( getLineReportedFlag( ) )
				    continue;
				else {
				    interpreter.applyForPrompt( );
				    break;
				}
			    }
			}
		    }.start( );
	    } else
		setLineReportedFlag( true );
	}
    }
}		      


