/***********************************************************************
 * Bdbj 1.2.1 (by Jon Cook August 2002)
 *
 * Interpreter.java 
 * 
 * Receive instructions from user interface in the first instance,
 * and deal with them if they relate to the entire debugging session,
 * otherwise pass them to the relevant thread(s).
 ***********************************************************************/

package jjc.bdbj.client;

import java.io.*;

public class Interpreter extends Thread {

    public static int MAX_THREADS = 128;

    public ClientSocket clientSocket = null;
    Breakpoints breakpoints = null;

    private SharedBuffer firstThreadNotify = new SharedBuffer( );
    private boolean notifiedFirstThread = false;
    
    private boolean prompt;
    private String classname;   public String getClassname( ) { return classname; }
    private int initialLine;    public int getInitialLine( ) { return initialLine; }
    private String initialFile; public String getInitialFile( ) { return initialFile; }
    private String fileList;    public String getDebugFiles( ) { return fileList; }
    private int bufferSize;     public int getBufferSize( ) { return bufferSize; }
    private String deepLogList; public String getDeepLogList( ) { return deepLogList; }

    public volatile DebugThread[] debugThreads = new DebugThread[ MAX_THREADS ];
    private DebugThread currentThread = null;

    private boolean promptPending = true; /* first prompt is before any step command */
    private synchronized void setPromptPending( ) {
	promptPending = true;
    }
    
    public synchronized void applyForPrompt( ) {
	if( prompt && promptPending ) {
	    System.out.print( "\nBdbj >" );
	    promptPending = false;
	}
    }
    
    public boolean amICurrentThread( DebugThread me ) {
	return me == currentThread;
    }

    public void newThread( String threadName, int thr, int shouldCreate ) throws NotEnoughMemoryException {
	if( ( shouldCreate != 0 ) && debugThreads[ thr ] == null ) {
	    long mem = Runtime.getRuntime( ).freeMemory( );
	    int bs = getBufferSize( );
            if( mem >= ( ( bs * 4 ) + ( 10 * 1024 ) ) )
		debugThreads[ thr ] = new DebugThread( this, clientSocket, fileList, 
						       prompt, thr, getBufferSize( ) );
	    else
		throw new NotEnoughMemoryException( );
	}
	debugThreads[ thr ].newThread( threadName );
	reportBirth( thr );
	if( !notifiedFirstThread ) {
	    firstThreadNotify.put( new Object( ) );
	    notifiedFirstThread = true;
	}
    }

    public void newMiniThread( int thr, int shouldCreate ) throws NotEnoughMemoryException {
	// only create the object array, nothing else.
	if( ( shouldCreate != 0 ) && debugThreads[ thr ] == null ) {
	    long mem = Runtime.getRuntime( ).freeMemory( );
	    int bs = getBufferSize( );
            if( mem >= ( ( bs * 4 ) + ( 10 * 1024 ) ) )		
		debugThreads[ thr ] = new DebugThread( this, clientSocket, fileList, 
						       prompt, thr, getBufferSize( ) );
	    else
		throw new NotEnoughMemoryException( );
	}
    }

    public void threadDead( int thr, int shouldDelete ) {
	debugThreads[ thr ].threadDead( );
	if( debugThreads[ thr ] == currentThread )
	    currentThread = null;
	if( shouldDelete != 0 )
	    debugThreads[ thr ] = null; // to allow the huge object array to be garbage collected.
    }

    public void miniThreadDead( int thr, int shouldDelete ) {
	if( shouldDelete != 0 ) 
	    debugThreads[ thr ] = null;
    }

    public void reportNewWatches( String watch ) {
        if( watch.indexOf( " == " ) == -1 )
            return; // this is a call intended for the old version
	clientSocket.writeLine( "watch " + watch );
    }

    public void reportNewLine( int line, int fileOffset, int thr, int gotoLine ) {
	debugThreads[ thr ].reportNewLine( line, fileOffset, gotoLine );
    }

    public void reportNewPc( int pc, String meth, int fileOffset, int thr, int gotoLine ) {
	debugThreads[ thr ].reportNewPc( pc, meth, fileOffset, gotoLine );
    }
    
    public void writePrompt( int thr ) {
	debugThreads[ thr ].writePrompt( );
    }
    
    public void cantBack( ) {
	clientSocket.writeLine( "error: Can't Step Back" );
    }  

    public int getCommand( int thr ) {
	return debugThreads[ thr ].getCommand( );
    }

    public String getWatches( int thr ) {
	return debugThreads[ thr ].getWatches( );
    }
    
    public String getNextBreakpointString( int thr ) {
	return debugThreads[ thr ].getBreakpointString( );
    }

    public void askFullThrow( Object exn, int thr ) {
	synchronized( clientSocket ) {
	    clientSocket.writeLine( "non-reversible throw on thread " + thr );
	    StringWriter sw = new StringWriter( );
	    ((Throwable)exn).printStackTrace( new PrintWriter( sw ) );
	    BufferedReader reader = new BufferedReader( new StringReader( sw.toString( ) ) );
	    String next;
	    try {
		while( ( next = reader.readLine( ) ) != null )
		    clientSocket.writeLine( "exception " + thr + ": " + next );
	    } catch( IOException _ ) { } 
	    clientSocket.writeLine( "end exception " + thr );
	}
    }

    public Interpreter( ClientSocket clientSocket, boolean prompt,
			String classname, int initialLine, String initialFile,
			String fileList, int bufferSize, String deepLogList,
			Breakpoints breakpoints ) {
	setName( "Interpreter" );
	setDaemon( true );

	Bridge.setInterpreter( this );

        this.clientSocket = clientSocket;
	this.prompt = prompt;
	this.classname = classname;
	this.initialLine = initialLine;
	this.initialFile = initialFile;
	this.fileList = fileList;
	this.bufferSize = bufferSize;
	this.deepLogList = deepLogList;
	this.breakpoints = breakpoints;

	for( int i = 0; i < MAX_THREADS; i++ )
	    debugThreads[ i ] = null;
	
	// for( int i = 0; i < 16; i++ ) 
	//    debugThreads[ i ] = new DebugThread( this, clientSocket, fileList, prompt, i );	

	Bridge.triggerInitialValueRequest( );
    }

    public void reportBirth( int threadNumber ) {
	if( currentThread == null )
	    currentThread = debugThreads[ threadNumber ];
    }

    private String[] helpText = {
    "Bdbj: Debugging mode:",
    "---------------------",
    "step             - step forwards one line",
    "stepover         - step over forwards one line",
    "back             - step backwards one line",
    "backover         - step over backwards one line",
    "bpt              - run to next breakpoint",
    "backbpt          - run to last breakpoint",
    "",
    "allstep          - all threads step",
    "allstepover      - all threads stepover",
    "allback          - all threads back",
    "allbackover      - all threads backover",
    "allbpt           - all threads to next breakpoint",
    "allbackbpt       - all threads to last breakpoint",
    "allotherbpt      - threads other than current run to breakpoint",
    "thread <integer> - change current thread",
    "",
    "bcstep, bcback,... - bytecode step, back, etc.",
    "",
    "watch <string>   - query the value of given variable",
    "toggle <integer> <string>  - toggle breakpoint status of line in file",
    "listbpts         - echo the table of breakpoints",
    "clearallbpts     - clear all breakpoints",
    "end/exit/quit/terminate - end debugging session",
    "help             - display this message",
    "input <string>   - send the line <string> to stdin",
    "gcbuffer         - resets the buffer, allowing garbage collection"
    };

    private void help( ) {
        for( int i = 0; i < helpText.length; i++ )
            clientSocket.writeLine( "> " + helpText[ i ] );
    }

    private void interpretCommand( String command ) {
        boolean writePrompt = true;
	if( command.startsWith( "+ " ) )
	    command = command.substring( 2 );
	if( command.equals( "" ) ) 
	    ;
	else if( command.startsWith( "thread" ) ) {
	    int thrNumber = Decoder.getNumber( command );
	    currentThread = debugThreads[ thrNumber ];
	}
        else if( command.equals( "step" ) || command.equals( "stepover" ) 
		 || command.equals( "back" ) || command.equals( "backover" )
		 || command.equals( "bpt" ) || command.equals( "backbpt" )
		 || command.startsWith( "watch" ) || command.startsWith( "bc" )
		 || command.equals( "gcbuffer" ) ) { 
	    setPromptPending( );
	    if( currentThread != null )
		currentThread.interpretCommand( command );
            writePrompt = false;
        }
	else if( command.startsWith( "step" ) )
	    try {
		debugThreads[ Decoder.getNumber( command ) ].interpretCommand( "step" );
	    } catch( Exception _ ) { }
	else if( command.startsWith( "back" ) ) {
	    setPromptPending( );
	    try {
		debugThreads[ Decoder.getNumber( command ) ].interpretCommand( "back" );
	    } catch( Exception _ ) { }
	    writePrompt = false;
	}
	else if( command.equals( "allotherbpt" ) ) {
	    for( int i = 0; i < MAX_THREADS; i++ )
		if( debugThreads[ i ] != null && debugThreads[ i ] != currentThread )
		    debugThreads[ i ].interpretCommand( "bpt" );
	    writePrompt = false;
	    if( prompt ) 
		new Thread( ) { 
			public void run( ) {
			    try {
				sleep( 200 );
			    } catch( InterruptedException _ ) {
			    }
			    System.out.print( "\nBdbj >" );
			}
		    }.start( );	    
        } 
	else if( command.startsWith( "all" ) ) {
	    setPromptPending( );
	    for( int i = 0; i < MAX_THREADS; i++ )
		if( debugThreads[ i ] != null )
		    debugThreads[ i ].interpretCommand( command.substring( 3 ) );
	    writePrompt = false;
	} 
	else if( command.startsWith( "toggle" ) ) {
	    try {
		String end = command.substring( 7 ).trim( );
		int pos = end.indexOf( ' ' );
		int line;
		String file;
		if( pos == -1 ) {
		    line = Integer.parseInt( end );
		    file = Debugger.getFilename( classname );
		} else {
		    line = Integer.parseInt( end.substring( 0, end.indexOf( ' ' ) ) );
		    file = end.substring( end.indexOf( ' ' ) + 1 ).trim( );
		}
		if( line < 0 )
		    clientSocket.writeLine( "error: Bad Line Number" );
		else
		    breakpoints.toggleBreakpoint( line, file );
	    } catch( Exception e ) {
		clientSocket.writeLine( "error: Bad Input" );
	    }
        }
	else if( command.equals( "listbpts" ) )
            breakpoints.listBreakpoints( );
        else if( command.equals( "clearallbpts" ) )
            breakpoints.clearAllBreakpoints( );
        else if( command.equals( "end" ) ||
                 command.equals( "quit" ) ||
                 command.equals( "exit" ) ||
                 command.equals( "terminate" ) ) {
            clientSocket.close( );
            System.exit( 0 );
        }
        else if( command.equals( "help" ) )
            help( );
        else
            clientSocket.writeLine( "error: Unrecognised Command : " + command );
        if( writePrompt && prompt ) {
	    System.out.print( "\nBdbj >" );
	}
    }

    public void run( ) {
	firstThreadNotify.get( );
        while( true ) {
            interpretCommand( clientSocket.readLine( ) );
	}
    }
}
