/***********************************************************************
 * Bdbj 1.2.1 (by Jon Cook August 2002)
 *
 * Debug.java 
 * 
 * Coordinate debugging: create the client process and oversee
 * communication with it via a socket.
 *
 * May 25-26 2002: fixed leak, threads not finishing when they should.
 * Aug 7 2002: fixed bug, first message from debugger lost because not in
 *             debug mode until after start message sent. 
 *
 ***********************************************************************/

package jjc.bdbj.interfaces.gui;

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.table.*;

public class Debug implements ActionListener, Runnable, MessageReceiver {

    ModeMenuBar menuBar = null;
    ModeToolBar toolBar = null;
    ModeToolBar fileToolBar = null;

    FileManager fileManager = null;
    ThreadManager threadManager = null;

    private Watches watches = null;
    private String classname = "";
    private String[] args = new String[] { };
    private String deepLoggingList = "java.util.*";
    static String kaffeOptions = System.getProperty( "bdbj.options" ).replace( '_', ' ' );;
    private int bufferSize = 128 * 1024;
    private ProcessSocket serverSocket = null;
    private boolean gcOn;

    public Debug( FileManager fileManager, ThreadManager threadManager, Watches watches,
		  String classname, String[] args, boolean gcOn ) {
	this.fileManager = fileManager;
	this.threadManager = threadManager;
	this.watches = watches;
        this.classname = classname;
        this.args = args;
	this.gcOn = gcOn;
	fileManager.setMessageReceiver( this );
	threadManager.setMessageReceiver( this );
	watches.setMessageReceiver( this );
    }

    public void setToolBar( ModeToolBar toolBar, ModeToolBar fileToolBar ) {
	this.toolBar = toolBar;
	this.fileToolBar = fileToolBar;
    }

    public void setMenuBar( ModeMenuBar menuBar ) {
        this.menuBar = menuBar;
    }

    private void debug( int init, String initFile, String script ) {
	if( fileManager.notCompiled( ) ) {
	    int option = DefaultDialog.yesNo( "Save and Compile?" );
	    if( option == JOptionPane.NO_OPTION ) 
		return;
	    else if( !fileManager.allSaveCompile( ) )
		return;
	}	   
	
        serverSocket = new ProcessSocket( classname, 
					  fileManager.getFilelist( ),
					  args, init, initFile, this, gcOn, bufferSize,
					  deepLoggingList, script );
	menuBar.changeMode( "Transition" );
	toolBar.changeMode( "Transition" );
	fileToolBar.changeMode( "Transition" );
        serverSocket.start( );
    }

    private void debug( ) {
        debug( 0, "", null );
    }

    private void debugMethod( ) {
        debug( fileManager.getCaretLine( ), fileManager.getCurrentFile( ), null );
    }

    private void runDebugScript( ) {
	debug( 0, "", fileManager.getScript( ) );
    }

    Vector executing = new Vector( );

    private void execute( ) {
        executing.addElement( new Execute( classname, args ) );   
    }

    private void stopExecuting( ) {
	for( Enumeration elts = executing.elements( );
	     elts.hasMoreElements( );
	     ) 
	    ( (Execute)elts.nextElement( ) ).destroy( );
	executing.removeAllElements( );
    }

    public void setArguments( String[] args ) {
	this.args = args;
    }

    private void editArguments( ) {
	new EditArguments( this, args ).setVisible( true );
    }

    private void changeBufferSize( ) {
	Object input = DefaultDialog.getInput( "Enter Buffer Size\n(roughly 8 buffer slots per line of code,\n"
					       + "each thread debugged has a buffer this size,\n"
					       + "default is 131072)\n",
					       "" + bufferSize );
	if( input != null ) {
	    try {
		int newBufferSize = Integer.parseInt( input.toString( ) );
		if( newBufferSize < 8 )
		    throw new NumberFormatException( );
		bufferSize = newBufferSize;
	    } catch( Exception e) {
		DefaultDialog.error( "Bad buffer size" );
	    }
	}
    }

    private void editDeepLoggingList( ) {
	Object input = DefaultDialog.getInput( "Enter the space separated list of classes, whose\n"
					     + "methods should be logged but not stepped through.\n"
					     + "A * wildcard may be used at the end of an item,\n"
					     + "e.g. * java.util.* javax.* java.io.File*\n",
					       "" + deepLoggingList );	
	if( input != null ) {
	    deepLoggingList = input.toString( );
	}
    }

    private void editKaffeOptions( ) {
	Object input = DefaultDialog.getInput( "Edit the options to be passed to Kaffe", kaffeOptions );
	if( input != null )
	    kaffeOptions = input.toString( );
    }

    private void showCommandConsole( ) {
	if( serverSocket != null )
	    if( serverSocket.commandConsole != null )
		serverSocket.commandConsole.setVisible( true );
    }
   
    public void deliverMessage( String message ) {
	if( message.equals( "stopexecute" ) ) {
	    stopExecuting( );
	    return;
	}
        if( serverSocket != null ) {
	    if( message.equals( "quit" ) ) {
		returnToEditing( );
		serverSocket.setDestroyedByDebugger( );
		serverSocket.destroy( );
	    } else 
		serverSocket.send( message );
	    if( message.startsWith( "thread" ) )
		watches.triggerFullWatchUpdate( );
	}
    }

    private String message;

    public synchronized void setMessage( String message ) {
	this.message = message;
    }

    private synchronized String getMessage( ) {
	return message;
    }


    int exnThread = -1;
    String exnMessage = "";
    
    public void run( ) {
	String message = getMessage( ).trim( );
	if( !menuBar.getCurrentMode( ).equals( "Debug" ) )
	    return;
	// if you change command names remember to change the values passed to substring.
        if( message.startsWith( "error:" ) )
            DefaultDialog.error( message.substring( 7 ) );
        else if( message.startsWith( "watch" ) )
            watches.reportWatch( message.substring( 6 ) );
	else if( message.startsWith( "new-thread" ) )
	    threadManager.newThread( message.substring( 11 ) );
	else if( message.startsWith( "thread-dead" ) )
	    threadManager.threadDead( message.substring( 12 ) );
        else if( message.startsWith( "line" ) || message.startsWith( "goto line" ) ) {
            try {
		boolean gotoLine = false; 
		if( message.startsWith( "goto" ) ) {
		    message = message.substring( 5 );
		    gotoLine = true;
		}
		int line = Integer.parseInt( message.substring( 5, message.indexOf( " in-file " ) ) );
		String file = message.substring( message.indexOf( " in-file " ) + 9,
						 message.indexOf( " on-thread " ) );
		String thrId = message.substring( message.indexOf( " on-thread " ) + 11 );
		int thr = Integer.parseInt( thrId.substring( 0, thrId.indexOf( ' ' ) ) );
                threadManager.changeCurrentLine( line, file, thr, gotoLine );
                watches.triggerFullWatchUpdate( );
            } catch( NumberFormatException _ ) { }
        }
	else if( message.startsWith( "pc" ) || message.startsWith( "goto pc" ) ) {
	    try {
		boolean gotoLine = false; 
		if( message.startsWith( "goto" ) ) {
		    message = message.substring( 5 );
		    gotoLine = true;
		}
		int pc = Integer.parseInt( message.substring( 3, message.indexOf( " in-method " ) ) );
		String method = message.substring( message.indexOf( " in-method " ) + 11,
						   message.indexOf( " in-file " ) );
		String file = message.substring( message.indexOf( " in-file " ) + 9,
						 message.indexOf( " on-thread " ) );
		String thrId = message.substring( message.indexOf( " on-thread " ) + 11 );
		int thr = Integer.parseInt( thrId.substring( 0, thrId.indexOf( ' ' ) ) );

                threadManager.changeCurrentPc( pc, method, file, thr, gotoLine );
                watches.triggerFullWatchUpdate( );
            } catch( NumberFormatException _ ) { }
	}
        else if( message.startsWith( "non-reversible throw on thread" ) ) {
	    try {
		exnThread = Integer.parseInt( message.substring( 31 ) );
		
	    } catch( NumberFormatException e ) { }
	} else if( message.startsWith( "exception" ) ) {
	    exnMessage += message.substring( message.indexOf( ":" ) + 1 ) + "\n"; 
	} else if( message.startsWith( "end exception" ) ) {
	    new NonReversibleThrowDialog( exnMessage, exnThread, this );
	    exnMessage = "";
	}
	else if( message.startsWith( ">" ) )
	    ;
	else
            DefaultDialog.error( "Internal error:\nBad message from BdbjClient" );
    }

    // called when the user commands the debugger to terminate
    public void returnToEditing( ) {
        menuBar.changeMode( "Edit" );
	toolBar.changeMode( "Edit" );
	fileToolBar.changeMode( "Edit" );
	fileManager.setMenuMode( "Edit" );
        // fileManager.setEditable( true );
    }

    private void resetCurrentLine2( ) {
	returnToEditing( ); // just in case exit was a System.exit( ) or uncaught exception.
	threadManager.removeCurrentLines( );
	threadManager.removeAllThreads( );
    }	

    // called when the subprocess exits
    public void resetCurrentLine( ) {
	try {
	    SwingUtilities.invokeLater( new Runnable( ) {
		    public void run( ) {
			resetCurrentLine2( );
		    }
		} );
	} catch( Exception e ) { }		       
    }

    public MenuData getMenu( ) {
	boolean[] justEdit = { true, false, false };
        boolean[] justDebug = { false, true, false };
        MenuItemData[] actionMenuData =
            { 
              new MenuItemData( "Debug", "Debug", this, "alt D", justEdit ),
              new MenuItemData( "Debug Method", "Debug Method", this, "shift alt D", justEdit ),
	      new MenuItemData( "Run Debug Script", "Run Debug Script", this, "", justEdit ),
	      new MenuItemData( "", "", null ),
              new MenuItemData( "Execute", "Execute", this, "alt E", justEdit ),
	      new MenuItemData( "Stop All Executing", "Stop All Executing", this, "alt shift E", justEdit ),
	      new MenuItemData( "", "", null ),
              new MenuItemData( "Edit Arguments", "Edit Arguments", this, "alt A", justEdit ),
	      new MenuItemData( "Change Buffer Size", "Change Buffer Size", this, "", justEdit ),
	      new MenuItemData( "Edit Deep Logging List", "Edit Deep Logging List", this, "", justEdit ),
	      new MenuItemData( "Edit Kaffe Options", "Edit Kaffe Options", this, "", justEdit ),
	      new MenuItemData( "", "", null ),
	      new MenuItemData( "Show Command Console", "Show Command Console", this, "", justDebug ),
	      new MenuItemData( "", "", null ),
	      new MenuItemData( "Step Over Backwards", "Step Over Backwards", this, "alt 1", justDebug ),
              new MenuItemData( "Step Backwards", "Step Backwards", this, "alt 2", justDebug ),
              new MenuItemData( "Step Forwards", "Step Forwards", this, "alt 3", justDebug ),
              new MenuItemData( "Step Over Forwards", "Step Over Forwards", this, "alt 4", justDebug ),
              new MenuItemData( "", "", null ),
              new MenuItemData( "Breakpoint Forwards", "Breakpoint Forwards", this, "alt 5", justDebug ),
	      new MenuItemData( "Breakpoint Backwards", "Breakpoint Backwards", this, "alt 6", justDebug ),
              new MenuItemData( "", "", null ),
              new MenuItemData( "Terminate", "Terminate", this, "alt 7", justDebug ),
              new MenuItemData( "Garbage Collect Buffer", "Garbage Collect Buffer", this, "", justDebug ) };
        return new MenuData( "Debug", actionMenuData );
    }

    public MenuData getThreadMenu( ) {
	boolean[] justDebug = { false, true, false };
	MenuItemData[] actionMenuData =
            { 
	      new MenuItemData( "All Step Over Backwards", "All Step Over Backwards", this, "", justDebug ),
              new MenuItemData( "All Step Backwards", "All Step Backwards", this, "", justDebug ),
              new MenuItemData( "All Step Forwards", "All Step Forwards", this, "", justDebug ),
              new MenuItemData( "All Step Over Forwards", "All Step Over Forwards", this, "", justDebug ),
              new MenuItemData( "", "", null ),
              new MenuItemData( "All Breakpoint Forwards", "All Breakpoint Forwards", this, "", justDebug ),
              new MenuItemData( "All Other Breakpoint Forwards", "All Other Breakpoint Forwards", 
				this, "", justDebug ), 
	      new MenuItemData( "All Breakpoint Backwards", "All Breakpoint Backwards", this, "", justDebug ) };

        return new MenuData( "Threads", actionMenuData );
    }

    public MenuData getByteCodeMenu( ) {
        boolean[] justDebug = { false, true, false };
        MenuItemData[] actionMenuData =
            { 
	      new MenuItemData( "Bytecode Step Over Backwards", "Bytecode Step Over Backwards", this, "alt ctrl 1", justDebug ),
              new MenuItemData( "Bytecode Step Backwards", "Bytecode Step Backwards", this, "alt ctrl 2", justDebug ),
              new MenuItemData( "Bytecode Step Forwards", "Bytecode Step Forwards", this, "alt ctrl 3", justDebug ),
              new MenuItemData( "Bytecode Step Over Forwards", "Bytecode Step Over Forwards", this, "alt ctrl 4", justDebug ),
              new MenuItemData( "", "", null ),
              new MenuItemData( "Bytecode Breakpoint Forwards", "Bytecode Breakpoint Forwards", this, "alt ctrl 5", justDebug ),
	      new MenuItemData( "Bytecode Breakpoint Backwards", "Bytecode Breakpoint Backwards", this, "alt ctrl 6", justDebug ) };
        return new MenuData( "Bytecode", actionMenuData );
    }

    public ToolBarData getDebugToolBar( ) {
	boolean[] justDebug = { false, true, false };
	boolean[] justEdit = { true, false, false };
	ToolBarButtonData[] debugToolBarData = 
	    {
		new ToolBarButtonData( "images/debug.gif", "Debug", this, justEdit ),
		new ToolBarButtonData( "", "", null ),
		new ToolBarButtonData( "images/backover.gif", "Step Over Backwards", this, justDebug ),
		new ToolBarButtonData( "images/back.gif", "Step Backwards", this, justDebug ),
		new ToolBarButtonData( "images/step.gif", "Step Forwards", this, justDebug ),
		new ToolBarButtonData( "images/stepover.gif", "Step Over Forwards", this, justDebug ),
		new ToolBarButtonData( "", "", null ),
		new ToolBarButtonData( "images/bpt.gif", "Breakpoint Forwards", this, justDebug ),
		new ToolBarButtonData( "images/backbpt.gif", "Breakpoint Backwards", this, justDebug ),
		new ToolBarButtonData( "", "", null ),
		new ToolBarButtonData( "images/terminate.gif", "Terminate", this, justDebug ) };
	return new ToolBarData( "Debug", debugToolBarData );
    }

    public void actionPerformed( ActionEvent event ) {
        String label = event.getActionCommand( );
        if( label.equals( "Step Over Backwards" ) ) 
            deliverMessage( "backover" );
        else if( label.equals( "Step Backwards" ) )
            deliverMessage( "back" );
        else if( label.equals( "Step Forwards" ) )
            deliverMessage( "step" );
        else if( label.equals( "Step Over Forwards" ) )
            deliverMessage( "stepover" );
        else if( label.equals( "Breakpoint Forwards" ) )
            deliverMessage( "bpt" );
	else if( label.equals( "Breakpoint Backwards" ) )
	    deliverMessage( "backbpt" );
	else if( label.equals( "Garbage Collect Buffer" ) )
	    deliverMessage( "gcbuffer" );
        else if( label.equals( "Terminate" ) ) {
	    returnToEditing( );
            deliverMessage( "quit" );
	} else if( label.equals( "All Step Over Backwards" ) )
	    deliverMessage( "allbackover" );
        else if( label.equals( "All Step Backwards" ) )
            deliverMessage( "allback" );
        else if( label.equals( "All Step Forwards" ) )
            deliverMessage( "allstep" );
        else if( label.equals( "All Step Over Forwards" ) )
            deliverMessage( "allstepover" );
        else if( label.equals( "All Breakpoint Forwards" ) )
            deliverMessage( "allbpt" );
	else if( label.equals( "All Other Breakpoint Forwards" ) )
	    deliverMessage( "allotherbpt" );
	else if( label.equals( "All Breakpoint Backwards" ) )
	    deliverMessage( "allbackbpt" );
        else if( label.equals( "Debug" ) )
            debug( );
        else if( label.equals( "Execute" ) )
            execute( );
	else if( label.equals( "Stop All Executing" ) )
	    stopExecuting( );
        else if( label.equals( "Debug Method" ) )
            debugMethod( );
	else if( label.equals( "Run Debug Script" ) )
	    runDebugScript( );
        else if( label.equals( "Edit Arguments" ) )
            editArguments( );
	else if( label.equals( "Change Buffer Size" ) )
	    changeBufferSize( );
	else if( label.equals( "Edit Deep Logging List" ) )
	    editDeepLoggingList( );
	else if( label.equals( "Edit Kaffe Options" ) )
	    editKaffeOptions( );
	else if( label.equals( "Show Command Console" ) )
	    showCommandConsole( );		
        else if( label.equals( "Bytecode Step Over Backwards" ) ) 
            deliverMessage( "bcbackover" );
        else if( label.equals( "Bytecode Step Backwards" ) )
            deliverMessage( "bcback" );
        else if( label.equals( "Bytecode Step Forwards" ) )
            deliverMessage( "bcstep" );
        else if( label.equals( "Bytecode Step Over Forwards" ) )
            deliverMessage( "bcstepover" );
        else if( label.equals( "Bytecode Breakpoint Forwards" ) )
            deliverMessage( "bcbpt" );
	else if( label.equals( "Bytecode Breakpoint Backwards" ) )
	    deliverMessage( "bcbackbpt" );
    }
}

class Execute {
    private String kaffeLocation = System.getProperty( "bdbj.kaffedir" );
    private String classPath = System.getProperty( "bdbj.classpath" );
    // private String kaffeOptions = System.getProperty( "bdbj.options" ).replace( '_', ' ' );

    private Process process = null;
    private Console console = null;
    private static String sep = System.getProperty( "file.separator" );

    private final SharedBuffer processTerminated = new SharedBuffer( );

    public void destroy( ) {
	if( process != null )
	    process.destroy( );
	if( console != null )
	    console.destroy( );
    }

    private String[] getStringArray( String str ) {
	str = str.trim( );
	String savedStr = str;
	int count = 0;
	for( ; ; ) {
	    int pos = str.indexOf( " " );
	    if( pos == -1 ) {
		if( str.length( ) != 0 ) 
		    count++;
		break;
	    }
	    count++;
	    str = str.substring( pos ).trim( );
	}
	str = savedStr;
	String[] strArray = new String[ count ];
	count = 0;
	for( ; ; ) {
	    int pos = str.indexOf( " " );
	    if( pos == -1 ) {
		if( str.length( ) != 0 )
		    strArray[ count ] = str;
		break;
	    }
	    strArray[ count ] = str.substring( 0, pos ).trim( );
	    count++;
	    str = str.substring( pos ).trim( );
	}
	return strArray;
    }

    private String[] vectorToStringArray( Vector vector ) {
	String[] result = new String[ vector.size( ) ];
	for( int i = 0; i < vector.size( ); i++ )
	    result[ i ] = vector.elementAt( i ).toString( );	
	return result;
    }

    public Execute( String classname, String[] args ) {
	if( kaffeLocation == null )
	    kaffeLocation = ".";
	if( classPath == null )
	    classPath = "";
	if( !classPath.equals( "" ) )
	    classPath = kaffeLocation + sep + "jre" + sep + "lib" + sep + "rt.jar:" + classPath;
	String kaffeOptions = Debug.kaffeOptions;
	if( kaffeOptions == null )
	    kaffeOptions = "";    
	final Vector argumentsVector = new Vector( );

	argumentsVector.add( kaffeLocation + sep + "bin" + sep + "kaffe" );

	if( !classPath.trim( ).equals( "" ) && ( kaffeOptions.indexOf( "-classpath " ) == -1 ) ) {
	    argumentsVector.add( "-classpath" );
	    argumentsVector.add( classPath );
	}
	if( !kaffeOptions.trim( ).equals( "" ) ) {
	    String[] options = getStringArray( kaffeOptions );
	    for( int i = 0; i < options.length; i++ )
		argumentsVector.add( options[ i ] );
	}

	argumentsVector.add( classname );

	for( int i = 0; i < args.length; i++ )
	    argumentsVector.add( args[ i ] );
		
	new Thread( ) {
		public void run( ) {
		    try {
			process = Runtime.getRuntime( ).exec( vectorToStringArray( argumentsVector ) );
		    } catch( IOException exception ) {
			return;
		    }
		    console = new Console( processTerminated, "stdin/stdout/stderr", 
					   process.getInputStream( ), process.getErrorStream( ), 
					   process.getOutputStream( ), true );
		    try {
			process.waitFor( );
		    } catch( InterruptedException _ ) { }
		    processTerminated.put( null );
		}
	    }.start( );
    }
}

class ProcessSocket extends Thread {
    public int DEFAULT_SOCKET = 24210;
    private static String sep = System.getProperty( "file.separator" );

    private Process process;
    private ServerSocket serverSocket = null;
    private Socket socket = null;

    private String kaffeLocation = System.getProperty( "bdbj.kaffedir" );
    private String classPath = System.getProperty( "bdbj.classpath" );
    // private String kaffeOptions = System.getProperty( "bdbj.options" ).replace( '_', ' ' );

    private String classname = "";
    private String[] filelist = null;
    private String[] args = null;
    private int bufferSize = 0;
    private int initial = 0;
    private String initFile = null;
    private Debug caller;
    private boolean gcOn;
    private String deepLoggingList = null;
    private String script = null;
    private String kaffeOptions = null;

    private boolean destroyedByDebugger = false;
    public void setDestroyedByDebugger( ) {
	destroyedByDebugger = true;
    }
    
    private SharedBuffer processTerminated_std = new SharedBuffer( );
    private SharedBuffer processTerminated_command = new SharedBuffer( );

    private PrintWriter socketOut = null;
    private static DataFromSocket socketIn = null;
    private /*static*/ Console console = null;
    public Console commandConsole = null;

    private PipedOutputStream pipedOutput = new PipedOutputStream( );
    private PipedInputStream pipedInput;
    private PrintWriter commandConsoleOutput = new PrintWriter( pipedOutput );

    private InputStream commandConsoleInput = null;
 

    public synchronized void send( String command ) {
        // synchronized for two reasons:
        // (i) socketOut is initialised by another thread
        // (ii) we don't want to allow two messages to get mixed together
	if( commandConsoleOutput != null ) {
	    commandConsoleOutput.println( "+ " + command );
	    commandConsoleOutput.flush( );
	}
        if( socketOut != null ) {
            socketOut.println( command );
            socketOut.flush( );
        }
    }

    private synchronized void setSocketOut( OutputStream outputStream ) {
	if( outputStream == null )
	    socketOut = null;
	else
	    socketOut = new PrintWriter( new OutputStreamWriter( outputStream ) );
    }

    public void destroy( ) {
	if( process != null )
	    process.destroy( );
    }

    public ProcessSocket( String classname, String[] filelist, 
			  String[] args, int initial, String initFile, Debug caller,
			  boolean gcOn, int bufferSize, String deepLoggingList,
			  String script ) {
        this.classname = classname;
	this.filelist = filelist;
        this.args = args;
	this.bufferSize = bufferSize;
        this.initial = initial;
	this.initFile = initFile;
        this.caller = caller;
	this.gcOn = gcOn;
	this.deepLoggingList = deepLoggingList;
	this.script = script;
	
	if( kaffeLocation == null )
	    kaffeLocation = ".";
	if( classPath == null || Debug.kaffeOptions.indexOf( "-classpath " ) != -1 )
	    classPath = "";
	if( !classPath.equals( "" ) )
	    classPath = kaffeLocation + sep + "jre" + sep + "lib" + sep + "rt.jar:" + classPath;
	kaffeOptions = Debug.kaffeOptions;
	if( kaffeOptions == null )
	    kaffeOptions = "";
	try {
	    pipedInput = new PipedInputStream( pipedOutput );
	} catch( Exception _ ) { }
    }

    public void run( ) {
	InputStream errorStream = null;
	InputStream inputStream = null;
	InputStream sockin = null;
        try {
	    while( true ) {
		if( socketIn != null) 
		    socketIn.interrupt( );
		if( console != null ) {
		    console.destroy( );
		    console = null;
		}

		// create the server socket
		try {
		    serverSocket = new ServerSocket( DEFAULT_SOCKET );
		} catch( Exception e ) {
		    serverSocket = new ServerSocket( 0 );
		}

		// create the subprocess
    
		process = Runtime.getRuntime( ).exec( kaffeLocation + sep + "bin" + sep + "kaffe -jar " 
						  + ( gcOn ? "" : "-nogcJJC " )
						  + ( classPath.trim( ).equals( "" ) ? "" : "-classpath " 
						      + classPath + " " )
						  + kaffeOptions + " -debugJJC "
						  + kaffeLocation + sep + "bdbj" + sep + "Bdbj.jar "
						  + serverSocket.getLocalPort( ) + " " 
						  + serverSocket.getInetAddress( ).getHostName( ) );
		
		// wait for connection, abort if takes longer than 1 minute

		serverSocket.setSoTimeout( 60 * 1000 ); 
		socket = serverSocket.accept( );
		
		( socketIn = new DataFromSocket( sockin = socket.getInputStream( ), pipedOutput, caller ) ).start( );
		setSocketOut( socket.getOutputStream( ) );				

		console = new Console( processTerminated_std, "stdin/stdout/stderr", 
				       inputStream = process.getInputStream( ), 
				       errorStream = process.getErrorStream( ), 
				       process.getOutputStream( ),
				       true );		

	        commandConsole = new Console( processTerminated_command, "Internal command console",
					      pipedInput, 
					      null,
					      socket.getOutputStream( ),
					      false );
		send( "class " + classname );
		for( int i = 0; i < filelist.length; i++ )
		    send( "file " + filelist[ i ] );
		
		boolean started = false;
		if( script == null ) {
		    for( int i = 0; i < args.length; i++ )
			send( "arg " + args[ i ] );
		    if( initial != 0 )
			send( "init " + initial + " " + initFile );
		    send( "buffer " + bufferSize );
		    send( "deeplog " + deepLoggingList );
		    
		    // the "Transition" menu bar mode is a hack designed to prevent the
		    // breakpoints being changed while the list is being sent to the
		    // client - the result of which is disastrous.
		    
		    caller.fileManager.sendAllBreakpoints( );
		    caller.menuBar.changeMode( "Debug" );
		    caller.toolBar.changeMode( "Debug" );
		    caller.fileToolBar.changeMode( "Debug" );
		    caller.fileManager.setMenuMode( "Debug" );
		    send( "start" );
		    started = true;
		} else {
		    BufferedReader br = new BufferedReader( new StringReader( script ) );
		    String next;
		    send( "buffer " + bufferSize );
		    send( "deeplog " + deepLoggingList );
		    while( ( next = br.readLine( ) ) != null ) {
			if( next.equals( "start" ) ) {
			    caller.fileManager.sendAllBreakpoints( );
			    caller.menuBar.changeMode( "Debug" );
			    caller.toolBar.changeMode( "Debug" );
			    caller.fileToolBar.changeMode( "Debug" );
			    caller.fileManager.setMenuMode( "Debug" );
			    started = true;
			}
			send( next );		     
		    }		   
		}

		if( !started ) {
		    caller.fileManager.sendAllBreakpoints( );
		    caller.menuBar.changeMode( "Debug" );
		    caller.toolBar.changeMode( "Debug" );
		    caller.fileToolBar.changeMode( "Debug" );
		    caller.fileManager.setMenuMode( "Debug" );
		    send( "start" );
		}
			       
		final int err = process.waitFor( );
		if( err != 0 && !destroyedByDebugger )
		    try {
			SwingUtilities.invokeLater( new Runnable( ) {
				public void run( ) {
				    DefaultDialog.info( "Process Terminated Abnormally:\nExit code " + err );
				}
			    } );
		    } catch( Exception e ) { }
		
		break;		    
	    }
        } catch( Exception e ) {
            System.out.println( "Exception with socket: " + e );
        } finally {	    
	    processTerminated_std.put( null );
	    processTerminated_command.put( null );
	    if( commandConsole != null )
		commandConsole.destroyIfNotVisible( );
	    try {
		pipedOutput.close( );
	    } catch( Exception _ ) { }
            try {
		setSocketOut( null );
                serverSocket.close( );
	    } catch( Exception _ ) { }
	    try {
                socket.close( );
	    } catch( Exception _ ) { }
	    try {
		sockin.close( );
            } catch( Exception _ ) { }
	    if( destroyedByDebugger ) {
		// sometimes the user will destroy the session because the target is spewing out
		// a huge amount;  so we cut off its output streams as well.
		// we don't do this when the target terminates normally, since then when the
		// user issued a BPT command with no breakpoints set, the output could be truncated.
		try {
		    inputStream.close( );
		} catch( Exception _ ) { }
		try {
		    errorStream.close( );
		} catch( Exception _ ) { }		
	    }
	    caller.resetCurrentLine( );
        }
    }
}

class DataFromSocket extends Thread {

    private Debug debug = null;
    private InputStreamReader socketInputStream = null;
    private PrintWriter commandConsoleOutput = null;
    
    public DataFromSocket( InputStream socketInputStream, OutputStream commandConsoleOutput, Debug debug ) {
        this.socketInputStream = new InputStreamReader( socketInputStream );
	this.commandConsoleOutput = new PrintWriter( commandConsoleOutput );
        this.debug = debug;
    }

    public void run( ) {
	String buffer = ""; 
	for( ; ; ) {
	    try {
		int data = socketInputStream.read( );
		if( data == -1 )
		    break;
		if( data == (int)'\n' ) {
		    
		    // Depending on the message, we call methods in either
		    // BreakpointFile or Watches.  
		    
		    // we must use invokeAndWait rather than invokeLater
		    // otherwise if the message load gets too high (as
		    // it does when there are lots of watches) we would
		    // begin to lose messages.
		    
		    commandConsoleOutput.println( "- " + buffer );
		    commandConsoleOutput.flush( );
		    debug.setMessage( buffer );
		    SwingUtilities.invokeAndWait( debug );
		    buffer = "";
		} else 
		    buffer += (char)data;
	    } catch( Exception _ ) { 
		break;
	    }
	}
    }
}

class Console implements ActionListener, WindowListener {
    private ConsoleArea area = null;
    private JFrame frame = null;
    private boolean std = false;
    
    private InputStream stdout, stderr;
    private OutputStream stdin;
    private SharedBuffer processTerminated;

    private boolean destroyed = false;
    //private synchronized boolean getDestroyed( ) { return destroyed; }
    //private synchronized boolean setDestroyed( ) { destroyed = true; }

    public void windowClosing( WindowEvent event ) {
        destroy( );
    }

    public void destroyIfNotVisible( ) {
	if( !frame.isVisible( ) )
	    destroy( );
    }

    public void windowIconified( WindowEvent event ) { }
    public void windowOpened( WindowEvent event ) { }
    public void windowDeiconified( WindowEvent event ) { }
    public void windowDeactivated( WindowEvent event ) { }
    public void windowClosed( WindowEvent event ) { }
    public void windowActivated( WindowEvent event ) { }

    public synchronized void destroy( ) {
	actionPerformed( null );
    }

    public String getText( ) {
	return area.getTextSynch( );
    }

    public void setVisible( boolean b ) {
	frame.setVisible( b );
    }

    public Console( SharedBuffer processTerminated, String title, InputStream stdout, InputStream stderr, OutputStream stdin, boolean std ) {
	this.processTerminated = processTerminated;
	this.std = std;
	this.stdout = stdout;
	this.stderr = stderr;
	this.stdin = stdin;

	area = new ConsoleArea( stdout, stderr, stdin, std );
	frame = new JFrame( title );
	
	frame.getContentPane( ).setLayout( new BorderLayout( ) );
	frame.getContentPane( ).add( "Center", area.getScrollPane( ) );
        JButton button = new JButton( "Dismiss" );
        frame.getContentPane( ).add( "South", button );
        button.addActionListener( this );
	area.startThreads( );
	frame.pack( );
	frame.setVisible( std );
    }

    public void actionPerformed( ActionEvent e ) {
	synchronized( this ) {
	    if( !destroyed ) {
		destroyed = true;

		frame.setVisible( false );
		frame.dispose( );
		area.destroy( );
		new Thread( ) {
			public void run( ) {
			    processTerminated.get( ); // wait for process to finish, then close all its streams.
			    try {
				stdin.close( );
			    } catch( Exception _ ) { }
			    try {
				stdout.close( );
			    } catch( Exception _ ) { }
			    try {
				if( stderr != null )
				    stderr.close( );       	
			    } catch( Exception _ ) { }
			}
		    }.start( );
	    }
	}
    }
}

class ConsoleArea extends JTextArea implements KeyListener, Runnable {
    private JScrollPane scrollPane = null;
    private ListenOutput stdout = null;
    private ListenOutput stderr = null;
    private PeriodicMoveCaretToEnd caretMover = null;
    private BufferedWriter stdin = null;
    private boolean std;
    private boolean stdinClosed = false;

    public JScrollPane getScrollPane( ) {
	return scrollPane;
    }

    public void startThreads( ) {
	stdout.start( );
	if( stderr != null )
	    stderr.start( );
	caretMover.start( );
    }

    public void destroy( ) {
	caretMover.interrupt( );
	stdout.dismiss( );
	if( stderr != null )
	    stderr.dismiss( );
    }

    public ConsoleArea( InputStream stdout, InputStream stderr, OutputStream stdin, boolean std ) {
	super( 12, 50 );
	this.stdout = new ListenOutput( stdout );
	if( stderr != null )
	    this.stderr = new ListenOutput( stderr );
	this.caretMover = new PeriodicMoveCaretToEnd( );
	this.stdin = new BufferedWriter( new OutputStreamWriter( stdin ) );
	this.std = std;
	setFont( new Font( "Monospaced", Font.PLAIN, 14 ) );
	scrollPane = new JScrollPane( this );
	addKeyListener( this );
    }

    public String getTextSynch( ) {
	return getText( );
    }

    String currentLine = "";

    public void keyPressed( KeyEvent e ) { 
	try {
	    if( e.getKeyChar( ) == (char)4 ) {
		try {
		    stdinClosed = true;
		    stdin.close( );		   
		} catch( Exception _ ) { }		
	    }
	    if( ( !e.isActionKey( ) && 
		  !Character.isISOControl( e.getKeyChar( ) ) &&
		  e.getKeyChar( ) != KeyEvent.CHAR_UNDEFINED ) 
		|| e.getKeyCode( ) == KeyEvent.VK_DELETE )
		setCaretPosition( getText( ).length( ) );
	    if( e.getKeyCode( ) == KeyEvent.VK_ENTER ) {
		setCaretPosition( getText( ).length( ) );		    
		if( !stdinClosed ) {
		    stdin.write( currentLine );
		    stdin.newLine( );
		    stdin.flush( );
		    currentLine = "";
		}
	    } else if( e.getKeyCode( ) == KeyEvent.VK_BACK_SPACE ) {
		setCaretPosition( getText( ).length( ) );
		if( !currentLine.equals( "" ) )
		    currentLine = currentLine.substring( 0, currentLine.length( ) - 1 );
	    } 
	} catch( IOException _ ) { }
    }

    public void keyReleased( KeyEvent e ) { }
    public void keyTyped( KeyEvent e ) { 
	if( !Character.isISOControl( e.getKeyChar( ) ) )
	    currentLine += e.getKeyChar( );       
	try {
	    if( e.getKeyChar( ) == 22 && !stdinClosed ) {
		JTextArea temp = new JTextArea( );
		temp.paste( );
		currentLine += temp.getText( );
		setCaretPosition( getText( ).length( ) );
		stdin.write( currentLine );
		stdin.flush( );
		currentLine = "";
	    }
	} catch( IOException _ ) { }
    }

    final int SNIP_TRIGGER = 150 * 500;
    final int SNIP_LENGTH = 100 * 500;

    public void run( ) {
	int oldLength = getText( ).length( );
	if( oldLength > SNIP_TRIGGER ) {
	    String newText = getText( ).substring( oldLength - SNIP_LENGTH );
	    int indn = newText.indexOf( '\n' );
	    int indr = newText.indexOf( '\r' );
	    if( indn == -1 && indr == -1 ) 
		setText( "[snip]" + newText );
	    else {
		int min = indn == -1 ? indr : indr == -1 ? indn : Math.min( indn, indr );
		setText( "[snip]\n"+ newText.substring( min ) );
	    }
	}
	setCaretPosition( getText( ).length( ) );
    }

    class PeriodicMoveCaretToEnd extends Thread {
	// with high load we do not want to overload Swing with such requests,
	// so we send them periodically every 50ms.

	boolean flag = false;
	synchronized boolean getFlag( ) { return flag; }
	synchronized void setFlag( boolean flag ) { this.flag = flag; } 

	public void run( ) {
	    for( ; ; ) {
		try {
		    sleep( 50 );
		} catch( InterruptedException e ) {
		    break;
		}
		if( isInterrupted( ) )
		    break;
		if( getFlag( ) ) {
		    setFlag( false );
		    try {
			SwingUtilities.invokeAndWait( ConsoleArea.this );
		    } catch( Exception e ) {
			break;
		    }
		}
	    }
	}
    }

    class ListenOutput extends Thread {       
	BufferedReader stream = null;
	public ListenOutput( InputStream stream ) {
	    this.stream = new BufferedReader( new InputStreamReader( stream ) );
	}	    
	private boolean dismiss = false;
	private synchronized boolean dismissed( ) { return dismiss; }
	private synchronized void dismiss( ) { dismiss = true; }

	public void run( ) {
	    // listen for input from Stdin 
	    try {
		while( true ) {
		    int data = stream.read( );
		    if( data == -1 )
			break;
		    else {
			if( isInterrupted( ) || dismissed( ) )
			    break;
			ConsoleArea.this.append( "" + (char)data );
			ConsoleArea.this.caretMover.setFlag( true );
		    }
		}
	    } catch( Exception _ ) { 
	    }
	}
    }
}

class NonReversibleThrowDialog extends JDialog implements ActionListener {
    MessageReceiver receiver = null;
    int thr;
    JTextArea textArea;

    public NonReversibleThrowDialog( String errorMsg, int thr, MessageReceiver receiver ) {
	super( GraphicalInterface.frame, "Non-reversible Throw", true );
	this.receiver = receiver;
	this.thr = thr;
	getContentPane( ).setLayout( new BorderLayout( ) );
        getContentPane( ).add( "North", new NonReversibleThrowMessageBox( ) );
        getContentPane( ).add( "South", new NonReversibleThrowButtons( this ) );
        getContentPane( ).add( "Center", new JScrollPane( textArea = new JTextArea( errorMsg ) ) );
	textArea.setFont( new Font( "Monospaced", Font.PLAIN, 14 ) );
	textArea.setEditable( false );
	pack( );
	setVisible( true );
    }

    public void actionPerformed( ActionEvent event ) { 
	String label = event.getActionCommand( );
        if( label.equals( "Dispatch" ) )
	    receiver.deliverMessage( "step " + thr );
	else if( label.equals( "Step Back" ) )
	    receiver.deliverMessage( "back " + thr );
	setVisible( false );
	dispose( );
    }
}

class NonReversibleThrowMessageBox extends JPanel {
    public NonReversibleThrowMessageBox( ) {
	setLayout( new BorderLayout( ) );
	JPanel intermediate = new JPanel( );
	intermediate.add( new NonReversibleThrowMessage( ) );	 
	add( "Center", intermediate );
    }
}

class NonReversibleThrowMessage extends JPanel {
    public NonReversibleThrowMessage( ) {
	setLayout( new GridLayout( 4, 1 ) );
	add( new JLabel( "An exception has occured which cannot be reversed" ) );
	add( new JLabel( "as it is not caught by a logged method." ) );
	add( new JLabel( "Do you wish to allow the exception to be dispatched," ) );
	add( new JLabel( "or to step back to just before it was thrown?" ) );
    }
}	

class NonReversibleThrowButtons extends JPanel {
    public NonReversibleThrowButtons( ActionListener listener ) {
        JButton dispatchButton = new JButton( "Dispatch" );
	JButton stepBackButton = new JButton( "Step Back" );
        add( dispatchButton );
        add( stepBackButton );
        dispatchButton.addActionListener( listener );
        stepBackButton.addActionListener( listener );
   }                                     
}                                   

class EditArguments extends JDialog implements ActionListener {
    private ArgumentsTable table;
    private Debug debug = null;

    public EditArguments( Debug debug, String[] args ) {
	super( GraphicalInterface.frame, "Edit Arguments" );
	this.debug = debug;
        getContentPane( ).setLayout( new BorderLayout( ) );
        getContentPane( ).add( "North", new JLabel( "Edit the arguments (one per line)" ) );
        getContentPane( ).add( "South", new EditArgumentsButton( this ) );
        getContentPane( ).add( "Center", new JScrollPane( table = new ArgumentsTable( args ) ) );
        setSize( 500, 300 );
    }

    public void actionPerformed( ActionEvent event ) {
        String label = event.getActionCommand( );
        if( label.equals( "More" ) )
            table.insertAtEnd( );
        else if( label.equals( "Insert" ) )
            table.insertRow( );
        else if( label.equals( "Delete" ) )
            table.removeRow( );
        else if( label.equals( "OK" ) ) {
            debug.setArguments( table.getArguments( ) );
            setVisible( false );
        }
        else if( label.equals( "Cancel" ) ) {
            setVisible( false );
        }
    }
}

class EditArgumentsButton extends JPanel {
    public EditArgumentsButton( ActionListener listener ) {
        JButton moreButton = new JButton( "More" ), 
                insertButton = new JButton( "Insert" ),
                deleteButton = new JButton( "Delete" ),
                okButton     = new JButton( "OK" ),
                cancelButton = new JButton( "Cancel" );

        add( moreButton );
        add( insertButton );
        add( deleteButton );
        add( okButton );
        add( cancelButton );
        moreButton.addActionListener( listener );
        insertButton.addActionListener( listener );
        deleteButton.addActionListener( listener );
        okButton.addActionListener( listener );
        cancelButton.addActionListener( listener );
   }                                     
}                                   

class ArgumentsTable extends JTable {
    private DefaultTableModel tableModel = null;
    private TableColumnModel columnModel = null;
    private TableColumn varColumn = null;
    private JTextField textField = null;    
    private Object[] nullRow = { null };

    public ArgumentsTable( String[] args ) {
	setModel( tableModel = new DefaultTableModel( args.length, 1 ) );
	for( int i = 0; i < args.length; i++ )
	    setValueAt( args[ i ], i, 0 );
        setFont( new Font( "Monospaced", Font.PLAIN, 14 ) );

        setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
        setColumnModel( columnModel = new DefaultTableColumnModel( ) );

        columnModel.addColumn( varColumn = new TableColumn( 0 ) );
        textField = new JTextField( );
        textField.setFont(new Font( "Monospaced", Font.PLAIN, 14 ) );
        varColumn.setCellEditor( new DefaultCellEditor( textField ) );
	
	// workaround for a bug in JDK 1.4
	tableModel.addRow( nullRow );
	editCellAt( getRowCount( ) - 1, 0 );
	tableModel.addRow( nullRow );
	editCellAt( getRowCount( ) - 1, 0 );
	try { editingStopped( new ChangeEvent( this ) ); } catch( Exception _ ) { }
	tableModel.removeRow( getRowCount( ) - 1 );
	try { editingStopped( new ChangeEvent( this ) ); } catch( Exception _ ) { }
	tableModel.removeRow( getRowCount( ) - 1 );
	try { editingStopped( new ChangeEvent( this ) ); } catch( Exception _ ) { }
    }

    public String[] getArguments( ) {
	try { editingStopped( new ChangeEvent( this ) ); } catch( Exception _ ) { }
        int rows = getRowCount( );
        String[] ret = new String[ rows ];
        for( int i = 0; i < rows; i++ ) {
           Object next = getValueAt( i, 0 );
           if( next == null )
               ret[ i ] = "";
           else
               ret[ i ] = next.toString( );
        }
        return ret;
    }

    public void insertRow( ) {
        int selectedRow = getSelectedRow( );
        if( selectedRow != -1 ) {
	    try { editingStopped( new ChangeEvent( this ) ); } catch( Exception _ ) { }
            tableModel.insertRow( selectedRow, nullRow );
	    clearSelection( );
	    setRowSelectionInterval( selectedRow, selectedRow );
	    revalidate( );
	    repaint( );
	}
    }

    public void insertAtEnd( ) {
	try { editingStopped( new ChangeEvent( this ) ); } catch( Exception _ ) { }
	clearSelection( );
	tableModel.addRow( nullRow );
        setRowSelectionInterval( getRowCount( ) - 1,
                                 getRowCount( ) - 1 );
        revalidate( );
        repaint( );
    }

    public void removeRow( ) { 
	try { editingStopped( new ChangeEvent( this ) ); } catch( Exception _ ) { }
        int[] selectedRows = getSelectedRows( );
        for( int i = selectedRows.length - 1; i >= 0; i-- ) 
            tableModel.removeRow( selectedRows[ i ] );
        if( selectedRows.length > 1 )
            clearSelection( );
        revalidate( );
        repaint( );
    }
}

class ConnectStreams extends Thread {
    // this thread is used to copy the client's stderr/stdout to this program's stdout.
    private InputStream input = null;
    private OutputStream output = null;

    public ConnectStreams( InputStream input, OutputStream output ) {
	this.input = input;
	this.output = output;
    }

    public void run( ) {
	int available;
	try {
	    for( ; ; ) {
		int data = input.read( );
		if( data == -1 )
		    break;
		output.write( data );
		output.flush( );
	    }
	} catch( Exception _ ) { }
    }
}
