/***********************************************************************
 * Bdbj 1.2.1 (by Jon Cook August 2002)
 *
 * FileManager.java 
 * 
 * In charge to the top half of the screen including the thread selector.
 * Maintain a set of files, file handling, breakpoints, and current lines
 * of execution.   Also must create and update status bar for each file.
 *
 * 26/5/2002: Fixed bug: problem calling javap with older versions of JDK.
 * 21/4/2003: Added feautre: ability to set font size of source code.
 *
 ***********************************************************************/

package jjc.bdbj.interfaces.gui;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.undo.*;

class StatusBar extends JTextField implements Runnable {
    private String classname = "";

    public StatusBar( ) {
	super( "Initializing..." );
	setEditable( false );
	setFont( new Font( "Monospaced", Font.PLAIN, 14 ) );
    }

    private JTextArea textArea = null;
    private int caret = 0;
    private boolean textChanged = false;
    private boolean compiled = false;
    private String menuMode = "Edit";
    private int line = 0;
    private SafelyGetLine safelyGetLine = new SafelyGetLine( );

    // these get methods are private because their only purpose
    // is to protect shared data which is only read from within
    // this class.

    private synchronized int getCaretLocation( ) { return caret; }
    private synchronized boolean getTextChanged( ) { return textChanged; }
    private synchronized boolean getCompiled( ) { return compiled; }
    private synchronized String getMenuMode( ) { return menuMode; }

    public synchronized void setTextCaret( int caret ) { this.caret = caret; }
    public synchronized void setTextChanged( boolean textChanged ) { this.textChanged = textChanged; }
    public synchronized void setCompiled( boolean compiled ) { this.compiled = compiled; }
    public synchronized void setMenuMode( String menuMode ) { this.menuMode = menuMode; }
    public synchronized void setTextArea( JTextArea textArea ) { this.textArea = textArea; }
    public synchronized void setClass( String classname ) { this.classname = classname; }

    private int getCaretLine( ) {
	try {
	    SwingUtilities.invokeAndWait( safelyGetLine );
	} catch( Exception e ) {
	    System.out.println( e );
	}
	return line;
    }

    private String getChangedMode( ) {
	return getTextChanged( ) 
	    ? ( getCompiled( ) 
		? "          " 
		: "  Modified" )
	    : ( getCompiled( )
		? "  Compiled" 
		: "Uncompiled" );
    }

    private String getMode( ) {
	return getMenuMode( ).equals( "Debug" ) ? "Debugging" : "Editing  ";
    }

    private String getCaretLineJustified( ) {
	int line = getCaretLine( );
	String lineString = new Integer( getCaretLine( ) ).toString( );
	String output = "";
	int offset = 5 - lineString.length( );
	for( int i = 0; i < offset; i++ )
	    output += " ";
	output += lineString;
	return output;
    }

    public void updateStatusBar( ) {
	String newText = 
	    ( "Bdbj " + getMode( ) + " : " + getChangedMode( ) + " - Line " +
	      getCaretLineJustified( ) + " - " + classname );
	if( !getText( ).equals( newText ) ) 
	    setText( newText );
    }

    class SafelyGetLine implements Runnable {
	public void run( ) {
	    try {
		line = textArea.getLineOfOffset( caret ) + 1;
	    } catch( BadLocationException e ) {
	    }
	}
    }

    public void run( ) {
	while( true ) {
	    updateStatusBar( );
	    try {
		Thread.currentThread( ).sleep( 50 );
	    } catch( InterruptedException ignored ) {
	    }
	}
    }
}

class SourceStatusPanel extends JPanel {
    public SourceStatusPanel( BreakpointFile source ) {
	super( );
	JScrollPane sourceScrollPane = new JScrollPane( source );
	source.setScrollPane( sourceScrollPane );
	setLayout( new BorderLayout( ) );
        add( sourceScrollPane, BorderLayout.CENTER );
	add( source.getStatusBar( ), BorderLayout.SOUTH );	
    }
}

class LoadingPanel extends JPanel {
    public LoadingPanel( ) {
	super( );
	setLayout( new BorderLayout( ) );
	JLabel b;
	add( b = new JLabel( "Loading files...", JLabel.CENTER ), BorderLayout.CENTER );
	b.setFont( new Font( "SansSerif", Font.PLAIN, 32 ) ); 
    }
}

class BlankPanel extends JPanel {
    public BlankPanel( ) {
	super( );
	setLayout( new BorderLayout( ) );
	JLabel b;
	add( b = new JLabel( "No files loaded.", JLabel.CENTER ), BorderLayout.CENTER );
	b.setFont( new Font( "SansSerif", Font.PLAIN, 32 ) ); 
    }
}

class SourceCode extends JTextArea implements ActionListener {
    private String searchString = "";

    public SourceCode( ) {
	super( 15, 80 );
        setEditable( true ); 
        setFont( new Font( "Monospaced", Font.PLAIN, 14 ) );
        setTabSize( 4 );
    }   
    
    public SourceCode( int fontSize ) {
	super( 15, 80 );
        setEditable( true ); 
        setFont( new Font( "Monospaced", Font.PLAIN, fontSize ) );
        setTabSize( 4 );
    }   
	
    public int getStartOfLine( int line ) {
	try {
	    return getLineStartOffset( line - 1 );
	} catch( BadLocationException e ) {
	}
	return 0;
    }

    public int getEndOfLine( int line ) {
	try {
	    return getLineEndOffset( line - 1 );
	} catch( BadLocationException e ) {
	}
	return 0;
    }

    public int getLineFromCaret( int caret ) {
	try {
	    return getLineOfOffset( caret ) + 1;
	} catch( BadLocationException e ) {
	}
	return 0;
    }

    public int getCaretLine( ) {
        return getLineFromCaret( getCaretPosition( ) );
    }

    public int getStartOfCaretLine( ) {
        return getStartOfLine( getCaretLine( ) );
    }

    public int getEndOfCaretLine( ) {
        return getEndOfLine( getCaretLine( ) );
    }

    public void registerMouseListener( MouseListener listener ) {
        addMouseListener( listener );
    }

    private void gotoLine( String linetext ) {
        int line = 0;
        try {
            line = Integer.parseInt( linetext.trim( ) );
        } catch( NumberFormatException e ) {
            DefaultDialog.error( "Invalid number" );
            return;
        }
        if( line < 1 ) {
            DefaultDialog.error( "Invalid number" );
            return;
        }
	requestFocus( );
	reallyCentreOnLine( line );
	setCaretPosition( getEndOfLine( line ) );
	moveCaretPosition( getStartOfLine( line ) );
    } 

    JScrollPane scrollPane = null;
    public void setScrollPane( JScrollPane scrollPane ) {
	this.scrollPane = scrollPane;
    }

    protected void reallyCentreOnLine( int line ) { // added 19/12/01
	int viewHeight = scrollPane.getSize( ).height;
	int currentx = scrollPane.getViewport( ).getViewPosition( ).x;
	int newCentre = line * getRowHeight( );
	int top = newCentre - viewHeight / 2;
	scrollPane.getViewport( ).setViewPosition( new Point( currentx, top ) );
	scrollPane.repaint( );
	setCaretPosition( getStartOfLine( line ) );
    }

    protected void centreOnLine( int line ) { // added 19/12/01
	int viewHeight = scrollPane.getSize( ).height;
	int currentx = scrollPane.getViewport( ).getViewPosition( ).x;
	int currentPos = scrollPane.getViewport( ).getViewPosition( ).y;
	int newCentre = line * getRowHeight( );
	int top = newCentre - viewHeight / 2;
	if( newCentre < currentPos || newCentre > currentPos + viewHeight ) {
	    top = Math.max( 0, top );
	    scrollPane.getViewport( ).setViewPosition( new Point( currentx, top ) );
	    scrollPane.repaint( );
	}
	setCaretPosition( getStartOfLine( line ) );	
    }

    private void search( String str ) {
        searchString = str;
        int pos = getText( ).indexOf( str, getCaretPosition( ) + 1 );
        if( pos == -1 )
            DefaultDialog.error( "Not found" );
        else {
            setCaretPosition( pos );
	    select( pos, pos + str.length( ) );
            requestFocus( );
        }
    }

    private void searchAgain( ) {
        if( !searchString.equals( "" ) )
            search( searchString );
    }

    public void actionPerformed( ActionEvent event ) {
        String label = event.getActionCommand( );
	if( label.equals( "Cut" ) )
	    cut( );
	else if( label.equals( "Copy" ) )
	    copy( );
	else if( label.equals( "Paste" ) )
	    paste( );
        else if( label.equals( "Goto Line" ) ) {
            Object input = DefaultDialog.getInput( "Enter Line Number", "" );
            if( input != null )
                gotoLine( input.toString( ) );
        }
        else if( label.equals( "Search" ) ) {
            Object input = DefaultDialog.getInput( "Enter String to Search For", searchString );
            if( input != null )
                search( input.toString( ) );
        }
        else if( label.equals( "Search Again" ) )
            searchAgain( );
    }
}

class SourceFile extends SourceCode 
    implements ActionListener, DocumentListener, CaretListener, KeyListener, UndoableEditListener {

    protected UndoManager undoManager = new UndoManager( );
    public void undoableEditHappened( UndoableEditEvent event ) {
	undoManager.addEdit( event.getEdit( ) );
	updateUndoRedoMenuItems( );
    }

    public void updateUndoRedoMenuItems( ) {
	if( menuBar != null ) {
	    menuBar.setEnabled( "Edit", "Undo", undoManager.canUndo( ) );
	    menuBar.setEnabled( "Edit", "Redo", undoManager.canRedo( ) );
	}
    }

    protected ModeMenuBar menuBar = null;
    public void setMenuBar( ModeMenuBar menuBar ) {
	this.menuBar = menuBar;
    }

    private StatusBar statusBar = new StatusBar( );
    public StatusBar getStatusBar( ) {
	return statusBar;
    }

    boolean textChanged = false;
    boolean compiled = false;
    protected String debugname;
    protected String filename;

    public SourceFile( String debugname, String filename ) {
        this.debugname = debugname;
	this.filename = filename;
	addCaretListener( this );
        getDocument( ).addDocumentListener( this );
	addKeyListener( this );
	statusBar.setClass( debugname );
	statusBar.setTextArea( this );
	new Thread( statusBar ).start( );
	undoManager.die( );
	getDocument( ).addUndoableEditListener( this );
    }

    public void setTextChanged( boolean newValue ) {
	textChanged = newValue;
	statusBar.setTextChanged( newValue );
    }
    
    public void setCompiled( boolean newValue ) {
	compiled = newValue;
	statusBar.setCompiled( newValue );
    }

    public void changedUpdate( DocumentEvent event ) {
    }

    public void keyPressed( KeyEvent event ) { }
    public void keyReleased( KeyEvent event ) { }
    public void keyTyped( KeyEvent event ) {
	if( event.getKeyChar( ) == '\n' || event.getKeyChar( ) == '\r' ) {
	    try {
		String text = getText( );
		int offset = getCaretPosition( );
		int line = getLineOfOffset( offset - 1 );
		int start = getLineStartOffset( line );
		int end = getLineEndOffset( line );
		int i;
		for( i = start;
		     i < end - 1 && Character.isWhitespace( text.charAt( i ) );
		     i++ )
		    ;
		insert( text.substring( start, i ), offset );
		setCaretPosition( offset + ( i - start ) );
	    } catch( BadLocationException e ) { }
	}
    }

    public void insertUpdate( DocumentEvent event ) {
	if( !textChanged ) { 
	    setTextChanged( true );
	    setCompiled( false );
	}
    } 

    public void removeUpdate( DocumentEvent event ) {
	if( !textChanged ) {
	    setTextChanged( true );
	    setCompiled( false );
	}
    }

    public void caretUpdate( CaretEvent event ) {
	statusBar.setTextCaret( getCaretPosition( ) );
    }

    // so that debug can tell the status bar that it has 
    // changed the menu mode.
    public void setMenuMode( String menuMode ) {
	statusBar.setMenuMode( menuMode );
    }

    private boolean fileExists( String filename ) {
        File file = new File( filename );
        return file.exists( );
    }

    private void makeBackup( String filename ) {
        File file = new File( filename );
        File dest = new File( filename + "~" );
        file.renameTo( dest );
    }

    private boolean saveFile( String code, String filename ) {
        // save code string to filename <filename>
        // return true iff successful
        makeBackup( filename );
        RandomAccessFile output = null;
        try {
            output = new RandomAccessFile( filename, "rw" );
            output.writeBytes( code );
        } catch ( Exception e ) {
            return false;
        } finally {
            try {
                output.close( );
            } catch( Exception _ ) {
            }
        }
        return true;
    }

    private boolean saveSource( ) {
        String source = getText( );
        if( !saveFile( source, filename ) ) {
            DefaultDialog.error( "Failed to save" ); 
            return false;
        } else 
            return true;
    }

    public boolean save( ) {
        if( !saveFile( getText( ), filename ) ) {
            DefaultDialog.error( "Failed to save" );
	    return false;
	}
        else {
	    setTextChanged( false );
            return true;
        }
    }

    public void saveBackup( ) {
        JFileChooser fileDialog = new JFileChooser( );
	fileDialog.setCurrentDirectory( new File( System.getProperty( "user.dir" ) ) );
        int ret = fileDialog.showSaveDialog( null );

        if( ret == JFileChooser.APPROVE_OPTION ) {
            if( !saveFile( getText( ), fileDialog.getSelectedFile( ).getPath( ) ) )
                DefaultDialog.error( "Failed to save file" );
        }
    }
    
    private void restoreBackup( ) {
	if( textChanged ) {
	    int option = DefaultDialog.yesNo( "You have not saved the current file.\n" +
                                            "Changes will be lost if you load a file.\n" +
                                            "Are you sure you want to?" );
	    if( option == JOptionPane.NO_OPTION ) 
		return;
	}
        JFileChooser fileDialog = new JFileChooser( );
	fileDialog.setCurrentDirectory( new File( System.getProperty( "user.dir" ) ) );
        int ret = fileDialog.showOpenDialog( null );

        if( ret == JFileChooser.APPROVE_OPTION ) {
            if( !loadFile( fileDialog.getSelectedFile( ).getPath( ), true ) )
                DefaultDialog.error( "Failed to load file" );
        }
    }

    public boolean loadFile( String filename, boolean letUndoManagerDie ) {
        RandomAccessFile input = null;
        String source = "";
        try {
            input = new RandomAccessFile( filename, "r" );
            int size = (int)input.length( );
            byte[] buffer = new byte[ size ];
            input.readFully( buffer, 0, size );
            source = new String( buffer );
        } catch ( Exception e ) {
            return false;
        } finally {
            try {
                input.close( );
            } catch( Exception _ ) {
            }
        }
        setText( source );
	if( letUndoManagerDie ) {
	    setCaretPosition( 0 );
	    setTextChanged( false );
	    setCompiled( true );
	    undoManager.die( );
	    updateUndoRedoMenuItems( );
	}
        return true;
    }

    public void initialLoadSource( ) {
	String classname = debugname.substring( 0, debugname.length( ) - 5 );
        RandomAccessFile input = null;
        String source = "";
        try {
            input = new RandomAccessFile( filename, "r" );
            int size = (int)input.length( );
            byte[] buffer = new byte[ size ];
            input.readFully( buffer, 0, size );
            source = new String( buffer );
        } catch ( Exception e ) {
            source = "public class " + classname + " {\n\tpublic static void main( String[] args ) {\n\t\t\n\t}\n}";
            setText( source );
            setCaretPosition( 62 + classname.length( ) );
	    // not strictly necessary to set these two values as
	    // it happens as a result of the call to setText,
	    // via a listener.  I just thought it would be
	    // nice to have it explicitly here.
	    setTextChanged( true );
	    setCompiled( false );
            return;
        } finally {
            try {
                input.close( );
            } catch( Exception _ ) {
            }
        }
        setText( source );
	setCaretPosition( 0 );
        requestFocus( );
	undoManager.die( );
        setTextChanged( false );
	setCompiled( true );
    }

    public void actionPerformed( ActionEvent event ) {
        String label = event.getActionCommand( );
	if( label.equals( "Undo" ) ) {
	    try {
		undoManager.undo( );
	    } catch( CannotUndoException _ ) { }
	    updateUndoRedoMenuItems( );	   
	}
	else if( label.equals( "Redo" ) ) {
	    try {
		undoManager.redo( );
	    } catch( CannotRedoException _ ) { }
	    updateUndoRedoMenuItems( );
	}
	else if( label.equals( "Save Backup" ) )
            saveBackup( );
        else if( label.equals( "Restore Backup" ) )
            restoreBackup( );
	else 
	    super.actionPerformed( event );
    }
    
    public String toString( ) {
	return debugname; 
    }
}

class CurrentLine {
    public int line = 0;
    public Object highlight = null;
    public boolean hidingBreakpoint = false;
    public boolean currentThread = false;

    public CurrentLine( int line, Object highlight, boolean hidingBreakpoint ) {
	this.line = line;
	this.highlight = highlight;
	this.hidingBreakpoint = hidingBreakpoint;
    }

    public String toString( ) {
	return line + " " + hidingBreakpoint;
    }
}

class BytecodeBreakpointFile extends BreakpointFile implements ActionListener, MouseListener,
    CaretListener, DocumentListener {

    private DefaultHighlighter.DefaultHighlightPainter bytecodeLinePainter = 
        new DefaultHighlighter.DefaultHighlightPainter( Color.yellow );
    
    public DefaultHighlighter.DefaultHighlightPainter getCurrentLinePainter( ) {
	return bytecodeLinePainter;
    }

    public BytecodeBreakpointFile( String debugname, String filename ) {
	super( debugname, filename );
	setEditable( false );
    }

    private int getPc( int fileLine ) {
	int start = getStartOfLine( fileLine );
	int end = getEndOfLine( fileLine );
	String line = getText( ).substring( start, end ).trim( );
	try {
	    line = line.substring( 0, line.indexOf( ' ' ) );
	    return Integer.parseInt( line );
	} catch( Exception _ ) {
	    return 0;
	}
    }

    private String mungeClassname( String in ) {
	return in.replace( '$', '.' );
    }

    private String getClassMethod( int fileLine, String fileName ) {
	int start = getEndOfLine( fileLine );
	String text = getText( ).substring( 0, start );
	int pos = text.lastIndexOf( "\nMethod " );
	if( pos == -1 ) {
	    return "";
	}
	else {
	    text = text.substring( pos + 8 );
	    text = text.substring( 0, text.indexOf( "\n" ) );
	    return mungeClassname( debugname ) + ":" + generateInternalSig( text );
	}
    }

    protected /*override*/ void toggleBreakpoint( int line ) {
        if( isBreakpoint( line ) )
            deleteBreakpoint( line );
        else 
            insertBreakpoint( line );
        if( menuBar.getCurrentMode( ).equals( "Debug" ) )
            messageReceiver.deliverMessage( "toggle " + getPc( line ) + " " + getClassMethod( line, debugname ) );
    }

    public /*override*/ void sendAllBreakpoints( ) {
        for( int i = 0; i < BREAKPOINTS; i++ )
            if( breakpoints[ i ] != -1 )
                messageReceiver.deliverMessage( "toggle " + getPc( breakpoints[ i ] ) + " " + 
						getClassMethod( breakpoints[ i ], debugname ) );
            else
                break;
    }

    private String toInternalMunge( String in ) {
	in = in.trim( );
	if( in.indexOf( '/' ) != -1 )
	    return in;
	else
	    return in.replace( '.', '/' );
    }

    private String generateInternalType( String lname, boolean includeVoid ) {
	lname = lname.trim( );
	if( lname.equals( "void" ) )
	    return includeVoid ? "V" : "";
	else if( lname.equals( "int" ) )
	    return "I";
	else if( lname.equals( "long" ) )
	    return "J";
	else if( lname.equals( "float" ) )
	    return "F";
	else if( lname.equals( "double" ) )
	    return "D";
	else if( lname.equals( "boolean" ) )
	    return "Z";
	else if( lname.equals( "byte" ) )
	    return "B";
	else if( lname.equals( "char" ) )
	    return "C";
	else if( lname.equals( "short" ) )
	    return "S";
	else {
	    return "L" + toInternalMunge( lname ) + ";"; 
	}
    }

    private String generateInternalTypeWithArrays( String lname, boolean includeVoid ) {
	String output = "";
	while( lname.endsWith( "[]" ) ) {
	    output += "[";
	    lname = lname.substring( 0, lname.length( ) - 2 );
	}
	return output + generateInternalType( lname, includeVoid );
    }

    private String generateInternalSig( String informalSig ) {
	if( informalSig.startsWith( "Method " ) )
	    informalSig = informalSig.substring( 7 );

	if( informalSig.startsWith( "static " ) )
	    return "<clinit>()V";

	int pos = informalSig.indexOf( ' ' );
	if( pos == -1 ) {
	    informalSig = "void <init>" + informalSig.substring( informalSig.indexOf( '(' ) );
	    pos = informalSig.indexOf( ' ' );
	}
	String retType = informalSig.substring( 0, pos );

	informalSig = informalSig.substring( pos + 1 );
	int posLbrac = informalSig.indexOf( '(' );
	int posRbrac = informalSig.indexOf( ')' );
	String methName = informalSig.substring( 0, posLbrac );
        String typeSig = informalSig.substring( posLbrac + 1, posRbrac );
        typeSig = stripSpaces( typeSig );
	
	String output = methName + "(";
	while( typeSig.length( ) > 0 ) {
	    int posComma = typeSig.indexOf( ',' );
	    if( posComma == -1 ) {
		output += generateInternalTypeWithArrays( typeSig, false );
		typeSig = "";
	    } else {
	        output += generateInternalTypeWithArrays( typeSig.substring( 0, posComma ), false );
		typeSig = typeSig.substring( posComma + 1 );
	    }
	}
	output += ")" + generateInternalTypeWithArrays( retType, true );	
	return output;
    }

    private String toInformalMunge( String in ) {
	if( in.indexOf( "." ) != -1 )
	    return in;
	else
	    return in.replace( '/', '.' );
    }

    private String processArgs( String args ) {
	try {
	    StringReader rdr = new StringReader( args );
	    int next;
	    int bracketCount;
	    String output = "";
	    while( ( next = rdr.read( ) ) != -1 ) {
		bracketCount = 0;
		while( (char)next == '[' ) {
		    bracketCount++;
		    next = rdr.read( );
		}
		switch( (char)next ) {			
		case 'V': output += "void"; break;
		case 'I': output += "int"; break;				
		case 'J': output += "long"; break;
		case 'F': output += "float"; break;
		case 'D': output += "double"; break;
		case 'Z': output += "boolean"; break;
		case 'B': output += "byte"; break;
		case 'C': output += "char"; break;
		case 'S': output += "short"; break;
		case 'L': {
		    String lname = "";
		    while( ( next = rdr.read( ) ) != ';' )
			lname += (char)next;
		    output += toInformalMunge( lname );
		    break;
		}
		default: output += "ERROR";
		}
		for( int i = 0; i < bracketCount; i++ )
		    output += "[]";
		output += ",";
	    }	
	    if( output.equals( "" ) )
		return output;
	    else
		return output.substring( 0, output.length( ) - 1 );	
	} catch( IOException _ ) {
	    return "ERROR";
	}
    }
    
    private String processMethSig( String methSig, String classname ) {
	int startOfSig = methSig.indexOf( "(" );
	int endOfArgs  = methSig.indexOf( ")" );
	String meth = methSig.substring( 0, startOfSig );
	String arg  = methSig.substring( startOfSig + 1, endOfArgs );
	String ret  = methSig.substring( endOfArgs + 1 );	
	if( meth.equals( "<init>" ) )
	    return classname.replace( '.', '$' ) + "(" + processArgs( arg ) + ")";
	else if( meth.equals( "<clinit>" ) )
	    return "static {}";
	else {
	    String leftRet, rightRet;
	    String processedRet = processArgs( ret );
	    int arrPos = processedRet.indexOf( '[' );
	    if( arrPos == -1 ) {
		leftRet = processedRet;
		rightRet = "";
	    } else {
		leftRet = processedRet.substring( 0, arrPos );
		rightRet = processedRet.substring( arrPos );
	    }
	    return leftRet + " " +
		meth + "(" +
		processArgs( arg ) + ")" +
		rightRet;
	}
    }
    
    public void addCurrentLineHighlight( int pc, String method, boolean current ) {
	int line = getLineFromPcMethod( pc, method );
	if( line != 0 )
	    super.addCurrentLineHighlight( line, current );
    }

    /* remove spaces between the brackets */
    private String stripSpaces( String in ) {
	int l = in.indexOf( '(' );
	int r = in.indexOf( ')' );
	if( l == -1 || r == -1 || l > r )
	    return in;
	String ls = in.substring( 0, l );
	String rs = in.substring( r );
	StringBuffer ms = new StringBuffer( in.substring( l, r ) );
	int pos;
	while( ( pos = ms.toString( ).indexOf( ' ' ) ) != -1 )
	    ms.deleteCharAt( pos );
	return ls + ms.toString( ) + rs;	    
    }

    private int getLineFromPcMethod( int pc, String method ) {
	try {
	    String classname = getDebugname( );
	    if( classname.endsWith( ".class" ) )
		classname = classname.substring( 0, classname.length( ) - 6 );
	    method = processMethSig( method, classname );
	    BufferedReader rdr = new BufferedReader( new StringReader( getText( ) ) );
	    String next;
	    int line = 0;
	    while( ( next = rdr.readLine( ) ) != null ) {
		next = next.trim( );
		line++;
		if( stripSpaces( next ).startsWith( "Method " + method ) ) {
		    while( ( next = rdr.readLine( ) ) != null ) {
			next = next.trim( );
			line++;
			if( next.startsWith( pc + " " ) ) {
			    return line;
			}
		    }
		    break;
		}
	    }	 
	} catch( IOException _ ) { }
	System.out.println( "Failed to locate pc " + pc + " method " + method );
	return 0;
    }

    public void jumpToCurrentLineHighlight( int pc, String method ) {
	int line = getLineFromPcMethod( pc, method );
	if( line != 0 )
	    super.jumpToCurrentLineHighlight( line );
    }

    public void removeCurrentLineHighlight( int pc, String method ) {
	int line = getLineFromPcMethod( pc, method );
	if( line != 0 )	
	    super.removeCurrentLineHighlight( line );
    }

    public void setEditable( boolean _ ) {
	super.setEditable( false );
    }

    public boolean save( ) {
	return false;
    }

    public void saveBackup( ) {
    }

    public void restoreBackup( ) {
    }

    public String getFilename( ) {
	return filename.replace( '$', '.' );
    }

    public boolean loadFile( String classname, boolean reloadSource ) {
	String source = javap( classname );
	if( source != null ) {
	    setText( source );
	    setCaretPosition( 0 );
	    setTextChanged( false );
	    setCompiled( true );
	}
	return source != null;
    }

    public void regenerate( ) {
	loadFile( debugname, true );
    }

    public void initialLoadSource( ) {
	loadFile( debugname, true );
    }

    private static String sep = System.getProperty( "file.separator" );

    private String javap( String classname ) {
	if( classname.endsWith( ".class" ) )
	    classname = classname.substring( 0, classname.length( ) - 6 );
	classname = classname.replace( '.', '$' );
	String classPath = System.getProperty( "bdbj.classpath" );
	String kaffeLocation = System.getProperty( "bdbj.kaffedir" );
	/*
	if( !classPath.equals( "" ) )
	    classPath = kaffeLocation + sep + "jre" + sep + "lib" + sep + "rt.jar:" + classPath;
	*/
        String sep = System.getProperty( "file.separator" );

        if( classPath == null )
	    classPath = "";
	
	if( !classPath.trim( ).equals( "" ) )
	    classPath = " -classpath " + classPath;
	else
	    classPath = "";

	String boot = kaffeLocation + sep + "jre" + sep + "lib" + sep + "rt.jar";
        Process process;
        try {
            process = Runtime.getRuntime().exec( "javap -c -bootclasspath " + boot + classPath + " " + classname );
 
            String next;
            StringBuffer data = new StringBuffer( "" );

	    BufferedReader inputStream_out = new BufferedReader( new InputStreamReader( process.getInputStream( ) ) );
	    while( ( next = inputStream_out.readLine( ) ) != null ) {
                data = data.append( next + "\n" );
	    }

	    BufferedReader inputStream_err = new BufferedReader( new InputStreamReader( process.getErrorStream( ) ) );
            while( ( next = inputStream_err.readLine( ) ) != null ) {
                data = data.append( next + "\n" );
	    }

            if( process.waitFor( ) == 0 ) 
		return data.toString( );
	    else {
		DefaultDialog.error( "Could not disassemble class file" );
		return null;
	    }
        } catch( IOException e ) {
            DefaultDialog.error( "I/O Exception" );
        } catch( InterruptedException e ) {
            DefaultDialog.error( "Compilation interrupted" );
        }
	return null;
    }
}

class BreakpointFile extends SourceFile implements ActionListener, MouseListener,
    CaretListener, DocumentListener {

    // this class also handles the painting of the current line of execution,
    // this is done because if the current line of execution coincides with a breakpoint
    // we have to temporarily remove the breakpoint highlight to make the current line
    // highlight visible.

    private int latestSourceLength = 0;

    public static int BREAKPOINTS = 15;
    protected int[] breakpoints = new int[ BREAKPOINTS + 1 ];
    private Object[] breakpointHighlights = new Object[ BREAKPOINTS + 1 ];

    private Vector currentLineHighlights = new Vector( );

    private DefaultHighlighter.DefaultHighlightPainter breakpointPainter = 
        new DefaultHighlighter.DefaultHighlightPainter( Color.cyan );

    private DefaultHighlighter.DefaultHighlightPainter currentLinePainter =
        new DefaultHighlighter.DefaultHighlightPainter( Color.green );

    public DefaultHighlighter.DefaultHighlightPainter getCurrentLinePainter( ) {
	return currentLinePainter;
    }

    private SourceCode sourceCode = null;
    private Highlighter highlighter = null;

    protected MessageReceiver messageReceiver = null;
    public void setMessageReceiver( MessageReceiver messageReceiver ) {
	this.messageReceiver = messageReceiver;
    }

    public BreakpointFile( String debugname, String filename ) {
	super( debugname, filename );
        highlighter = getHighlighter( );
        initBreakpoints( );
	getDocument( ).addDocumentListener( this );
	addMouseListener( this );
    }
    
    public String getDebugname( ) {
	return debugname;
    }

    public void setDebugname( String name ) {
	debugname = name;
    }

    public String getFilename( ) {
	return filename;
    }

    public String packageDetect( ) {
	try {
	    // look for a package declaration at the start of the file.       
	    StringReader sourceCode = new StringReader( getText( ) );
	    StreamTokenizer tokenizer = new StreamTokenizer( sourceCode );
	    // tokenizer.resetSyntax( );
	    tokenizer.slashSlashComments( true );
	    tokenizer.slashStarComments( true );
	    tokenizer.wordChars( 'a', 'z' );
	    tokenizer.wordChars( 'A', 'Z' );
	    tokenizer.wordChars( '_', '_' );
	    tokenizer.wordChars( '$', '$' );
	    tokenizer.wordChars( '.', '.' );
	    tokenizer.wordChars( '0', '9' );
	    int next;
	    while( ( next = tokenizer.nextToken( ) ) == StreamTokenizer.TT_EOL )
		;
	    if( next == StreamTokenizer.TT_WORD ) {
		if( tokenizer.sval.equals( "package" ) ) {
		    next = tokenizer.nextToken( );
		    return tokenizer.sval.replace( '.', '/' ) + "/";
		}
	    }
	} catch( IOException _ ) { }
	return "";
    }

    public synchronized void addCurrentLineHighlight( int line, boolean current ) {
	boolean bptFound = false;
	int toBeRemoved = 0;
	for( int i = 0; i < BREAKPOINTS; i++ ) {
	    if( breakpoints[ i ] == line ) {
		bptFound = true;
		toBeRemoved = i;
	    } else if( breakpoints[ i ] == -1 )
		break;
	}

	if( bptFound ) {
	    boolean alreadyHidden = false;
	    CurrentLine currentLine = null;
	    for( Enumeration elts = currentLineHighlights.elements( );
		 elts.hasMoreElements( );
		 ) {
		currentLine = (CurrentLine)elts.nextElement( );
		if( currentLine.line == line && currentLine.hidingBreakpoint )
		    alreadyHidden = true;
	    }	   
	    if( !alreadyHidden )
		highlighter.removeHighlight( breakpointHighlights[ toBeRemoved ] );
	}        

	Object highlight = null;
	try {
	    highlight = highlighter.addHighlight( getStartOfLine( line ),
						  getEndOfLine( line ),
						  getCurrentLinePainter( ) );
	} catch( BadLocationException e ) {}
	if( current ) {
	    centreOnLine( line );
	}
	currentLineHighlights.addElement( new CurrentLine( line, highlight, bptFound ) );
    }

    public void jumpToCurrentLineHighlight( int line ) {
	reallyCentreOnLine( line );
    }

    public synchronized void removeCurrentLineHighlight( int line ) {
	CurrentLine currentLine = null;
	for( Enumeration elts = currentLineHighlights.elements( );
	     elts.hasMoreElements( );
	     ) {
	    currentLine = (CurrentLine)elts.nextElement( );
	    if( currentLine.line == line ) {
		highlighter.removeHighlight( currentLine.highlight );

		if( currentLine.hidingBreakpoint ) {
		    int hidingCount = 0;
		    CurrentLine currentLine2 = null;
		    for( Enumeration elts2 = currentLineHighlights.elements( );
			 elts2.hasMoreElements( );
			 ) {
			currentLine2 = (CurrentLine)elts2.nextElement( );
			if( currentLine2.line == line && currentLine2.hidingBreakpoint )
			    hidingCount++;
		    }
		    if( hidingCount == 1 ) {
			for( int i = 0; i < BREAKPOINTS; i++ )
			    if( breakpoints[ i ] == line ) {
				try {
				    breakpointHighlights[ i ] =
					highlighter.addHighlight( getStartOfLine( line ),
								  getEndOfLine( line ),
								  breakpointPainter );
				} catch( BadLocationException _ ) { }
				break;
			}
		    }
		}
		
		currentLineHighlights.removeElement( currentLine );
		return;
	    }
	}
    }

    public synchronized void removeAllCurrentLineHighlights( ) {
	while( !currentLineHighlights.isEmpty( ) )
	    removeCurrentLineHighlight( ((CurrentLine)(currentLineHighlights.firstElement( ))).line );
    }

    protected boolean isBreakpoint( int line ) {
        for( int i = 0; i < BREAKPOINTS; i++ )
            if( breakpoints[ i ] == line ) 
                return true;
            else if( breakpoints[ i ] == -1 )
                return false;
        return false;
    }

    protected synchronized void insertBreakpoint( int line ) {
        // Crucially assumes the caret is currently on this line number
	
        // prevent the user from making an empty line a breakpoint (they are hard to see and the user
        // might not know where they are to remove them).	
        if( getEndOfCaretLine( ) - 1 ==
            getStartOfCaretLine( ) )
            return;

	if( FileManager.breakpointLimit( ) ) {
	    DefaultDialog.error( "Too Many Breakpoints" );
	    return;
	}

        for( int i = 0; i < BREAKPOINTS; i++ )
            if( breakpoints[ i ] == -1 ) {
                breakpoints[ i ] = line;
		boolean found = false;
		CurrentLine currentLine = null;
		for( Enumeration elts = currentLineHighlights.elements( );
		     elts.hasMoreElements( );
		     ) {
		    currentLine = (CurrentLine)elts.nextElement( );
		    if( currentLine.line == line ) {
			currentLine.hidingBreakpoint = true;
			found = true;
			break;
		    }
		}
		if( !found ) {
                    try {
                        breakpointHighlights[ i ] =  
                            highlighter.addHighlight( getStartOfCaretLine( ), 
                                                      getEndOfCaretLine( ), breakpointPainter );
                    } catch( Exception _ ) { }
                FileManager.addBreakpoint( );
		}
		return;
	    }
        DefaultDialog.error( "Too Many Breakpoints" );	
    }

    protected synchronized void deleteBreakpoint( int line ) {
        for( int i = 0; i < BREAKPOINTS; i++ )
            if( breakpoints[ i ] == line ) {
		
		boolean found = false;
		CurrentLine currentLine = null;
		for( Enumeration elts = currentLineHighlights.elements( );
		     elts.hasMoreElements( );
		     ) {
		    currentLine = (CurrentLine)elts.nextElement( );
		    if( currentLine.line == line ) {
			currentLine.hidingBreakpoint = false;
			found = true;
			break;
		    }
		}
		if( !found )
                    highlighter.removeHighlight( breakpointHighlights[ i ] );

                for( int j = i; j < BREAKPOINTS; j++ ) {
                    breakpoints[ j ] = breakpoints[ j + 1 ];
                    breakpointHighlights[ j ] = breakpointHighlights[ j + 1 ];
                }
		FileManager.removeBreakpoint( );
		return;
            }
    }

    public synchronized void clearAllBreakpoints( ) {
        while( breakpoints[ 0 ] != -1 )
            toggleBreakpoint( breakpoints[ 0 ] );
    }

    private synchronized void initBreakpoints ( ) {
        for( int i = 0; i < BREAKPOINTS + 1; i++ ) 
            breakpoints[ i ] = -1;
    }

    protected void toggleBreakpoint( int line ) {
        if( isBreakpoint( line ) )
            deleteBreakpoint( line );
        else 
            insertBreakpoint( line );
        if( menuBar.getCurrentMode( ).equals( "Debug" ) )
            messageReceiver.deliverMessage( "toggle " + line + " " + debugname );
    }

    public void breakpointHasBeenToggled( int line ) {
        if( isBreakpoint( line ) )
            deleteBreakpoint( line );
        else 
            insertBreakpoint( line );
    }

    public void sendAllBreakpoints( ) {
        for( int i = 0; i < BREAKPOINTS; i++ )
            if( breakpoints[ i ] != -1 )
                messageReceiver.deliverMessage( "toggle " + breakpoints[ i ] + " " + debugname );
            else
                break;
    }

    private void toggleBreakpoint( ) {
        toggleBreakpoint( getCaretLine( ) );
    }

    public void changedUpdate( DocumentEvent event ) {
	super.changedUpdate( event );
    }

    private void documentChanged( ) {
	int lines = getLineCount( );
	if( latestSourceLength == 0 )
	    latestSourceLength = lines;
	else if( lines != latestSourceLength ) {
	    clearAllBreakpoints( );
	    messageReceiver.deliverMessage( "quit" );
	    latestSourceLength = lines;
	}
    }

    public void insertUpdate( DocumentEvent event ) {
	super.insertUpdate( event );
	documentChanged( );
    } 

    public void removeUpdate( DocumentEvent event ) {
	super.removeUpdate( event );
	documentChanged( );
    }

    public void actionPerformed( ActionEvent event ) {
        String label = event.getActionCommand( );
        if( label.equals( "Toggle Breakpoint" ) )
            toggleBreakpoint( );
        else if( label.equals( "Clear All Breakpoints in File" ) )
            clearAllBreakpoints( );
	else
	    super.actionPerformed( event );
    }

    public void mouseClicked( MouseEvent event ) {
        if( event.isControlDown( ) && !menuBar.getCurrentMode( ).equals( "Transition" ) ) 
            toggleBreakpoint( );
    }

    public void mouseEntered( MouseEvent event ) {}
    public void mouseExited( MouseEvent event ) {}
    public void mousePressed( MouseEvent event ) {}
    public void mouseReleased( MouseEvent event ) {}
}

public class FileManager extends Vector implements ActionListener, WindowListener,
                                                   ListSelectionListener, Runnable {
    private FileCards fileCards = new FileCards( );
    private BreakpointFile file = null; /* the currently selected file */
    private JList fileList = null;
    private JList threadList = null;
    private String mainClass = null;
    private boolean editable = true;
    private String menuMode = "";
    private ThreadManager threadManager = null;
    String[] debugFiles;
    JFrame frame;
    private int fontSize = 14;

    static int totalNumberOfBreakpoints = 0; // we don't let this get above 15, synchronize on the class.
    static synchronized boolean breakpointLimit( ) {
	return totalNumberOfBreakpoints == BreakpointFile.BREAKPOINTS;
    }
    static synchronized void addBreakpoint( ) {
	totalNumberOfBreakpoints++;
    }
    static synchronized void removeBreakpoint( ) {
	totalNumberOfBreakpoints--;
    }

    private String stripClassname( String filename ) {
	String separator = System.getProperty( "file.separator" );
	String end;
	int lastSep = filename.lastIndexOf( separator );
	if( lastSep == -1 )
	    end = filename;
	else
	    end = filename.substring( lastSep + separator.length( ) );
	return end;
    }

    public void run( ) {
	loadExtraFiles( debugFiles );
    }
    
    public FileManager( String classname, String[] debugFiles, ThreadManager threadManager,
			JFrame frame ) {
	super( 10 );
	this.mainClass = classname;
	this.debugFiles = debugFiles;       
	this.threadManager = threadManager;
	this.frame = frame;
	threadList = threadManager.getThreadList( );
	threadManager.setFileManager( this );
	
   	fileCards.add( new LoadingPanel( ), "!loading" );
	fileCards.add( new BlankPanel( ), "!null" );
	
	if( debugFiles.length == 0 ) {
	    addElement( file = new BreakpointFile( mainClass + ".java", mainClass + ".java" ) );
	    file.initialLoadSource( );
	    SourceStatusPanel ssp = new SourceStatusPanel( file );
	    fileCards.add( ssp, mainClass + ".java" );
	    fileCards.showCard( mainClass + ".java" );
	} else 
	    fileCards.showCard( "!loading" );

	fileList = new JList( this );
	fileList.addListSelectionListener( this );

	fileList.setSelectedIndex( 0 );
    }

    public void loadExtraFiles( String[] debugFiles ) {
	for( int i = 0; i < debugFiles.length; i++ ) {
	    realOpenFile( debugFiles[ i ] );
	}
	fileList.setSelectedIndex( 0 );
    }

    public String[] getFilelist( ) {
	String[] filelist = new String[ size( ) ];
	for( int i = 0; i < size( ); i++ )
	    filelist[ i ] = elementAt( i ).toString( ).replace( '$', '.' );
	return filelist;
    }

    public void valueChanged( ListSelectionEvent event ) {
	if( fileList.getModel( ).getSize( ) == 0 )
	    return;
	Object o = fileList.getSelectedValue( );
	if( o != null ) {
	    BreakpointFile selected = (BreakpointFile)o;
	    fileCards.showCard( selected.getDebugname( ) );
	    file = selected;
	    file.updateUndoRedoMenuItems( );
	}
    }

    private MessageReceiver messageReceiver = null;
    public void setMessageReceiver( MessageReceiver messageReceiver ) {
	this.messageReceiver = messageReceiver;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ( (BreakpointFile)elts.nextElement( ) ).setMessageReceiver( messageReceiver ) )
	    ; // empty
    }

    private ModeMenuBar menuBar = null;
    public void setMenuBar( ModeMenuBar menuBar ) {
	this.menuBar = menuBar;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ( (BreakpointFile)elts.nextElement( ) ).setMenuBar( menuBar ) )
	    ; // empty
    }

    private Component watchTablePane = null;
    public void setWatchTablePane( Component watchTablePane ) {
	this.watchTablePane = watchTablePane;
    }
	
    public int getCaretLine( ) {
	return file.getCaretLine( );
    }

    public String getCurrentFile( ) {
	return file.getDebugname( );
    }

    public void setEditable( boolean b ) {
	editable = b;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ( (BreakpointFile)elts.nextElement( ) ).setEditable( b ) )
	    ; // empty
    }

    public void updateUndoRedoMenuItems( ) {
	if( file != null )
	    file.updateUndoRedoMenuItems( );
    }    

    public void clearAllBreakpoints( ) {
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ( (BreakpointFile)elts.nextElement( ) ).clearAllBreakpoints( ) )
	    ; // empty
    }

    public void toggleLineFile( int line, String file ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.getDebugname( ).equals( file ) )
		next.breakpointHasBeenToggled( line );
	}
    }	

    public void addThreadHighlight( int line, String file, boolean current, boolean gotoLine ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( ); 
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.getDebugname( ).equals( file ) ) {
		if( ( gotoLine && current ) && ( next != this.file ) ) {
		    fileCards.showCard( next.getDebugname( ) );
		    fileList.setSelectedValue( next, true );
		}
		next.addCurrentLineHighlight( line, current );
	    }
	}
    }

    public void addThreadHighlight( int pc, String meth, String file, boolean current, boolean gotoLine ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( ); 
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.getDebugname( ).equals( file ) ) {
		if( ( gotoLine && current ) && ( next != this.file ) ) {
		    fileCards.showCard( next.getDebugname( ) );
		    fileList.setSelectedValue( next, true );
		}
		((BytecodeBreakpointFile)next).addCurrentLineHighlight( pc, meth, current );
	    }
	}
    }

    public void jumpToThreadHighlight( int line, String file ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( ); 
	     ) {
	    next = (BreakpointFile)elts.nextElement( );

	    if( next.getDebugname( ).equals( file ) ) {
		if( ( next != this.file ) ) {
		    fileCards.showCard( next.getDebugname( ) );
		    fileList.setSelectedValue( next, true );
		}
		next.jumpToCurrentLineHighlight( line );
	    }
	}
    }

    public void jumpToThreadHighlight( int pc, String method, String file ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( ); 
	     ) {
	    next = (BreakpointFile)elts.nextElement( );

	    if( next.getDebugname( ).equals( file ) ) {
		if( ( next != this.file ) ) {
		    fileCards.showCard( next.getDebugname( ) );
		    fileList.setSelectedValue( next, true );
		}
		((BytecodeBreakpointFile)next).jumpToCurrentLineHighlight( pc, method );
	    }
	}
    }

    public void removeThreadHighlight( int line, String file ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.getDebugname( ).equals( file ) )
		next.removeCurrentLineHighlight( line );
	}
    }
    
    public void removeThreadHighlight( int pc, String method, String file ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.getDebugname( ).equals( file ) )
		((BytecodeBreakpointFile)next).removeCurrentLineHighlight( pc, method );
	}
    }

    public void setMenuMode( String newMode ) {
	menuMode = newMode;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ( (BreakpointFile)elts.nextElement( ) ).setMenuMode( newMode ) )
	    ; // empty
    }

    public void sendAllBreakpoints( ) {
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ( (BreakpointFile)elts.nextElement( ) ).sendAllBreakpoints( ) )
	    ; // empty
    }

    public String getScript( ) {
	if( file != null )
	    if( file.getFilename( ).endsWith( ".bdbj" ) )
		return file.getText( );
	
	BreakpointFile next = null;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( ); ) {
	    next = ( (BreakpointFile)elts.nextElement( ) );
	    if( next.getFilename( ).endsWith( ".bdbj" ) )
		return next.getText( );
	}
	DefaultDialog.error( "Cannot find a script.\nThe script filename must end\nwith .bdbj\n" );
	return null;
    }

    public JPanel getFileManagerPanel( ) {
	/* Mostly sticking panes inside panes */
	JPanel fileListPanel = new ListPanel( "Source files", fileList );
	JPanel threadListPanel = new ListPanel( "Threads", threadList );

	JPanel leftPanel = new JPanel( );
	leftPanel.setLayout( new GridLayout( 2, 1 ) );
	leftPanel.add( new JScrollPane( fileListPanel ) ); 
	leftPanel.add( new JScrollPane( threadListPanel ) );
	
	int preferredHeight = leftPanel.getPreferredSize( ).height;

	JSplitPane splitPane = new JSplitPane( JSplitPane.HORIZONTAL_SPLIT, true,
                                               leftPanel,
					       fileCards );
	/* Now, stop the UI doing something stupid to my splitPane when the user tries to resize the window */
	splitPane.setDividerLocation( fileListPanel.getPreferredSize( ).width + 5 );

	JPanel fullPanel = new JPanel( );
	fullPanel.setLayout( new BorderLayout( ) );
	fullPanel.add( "Center", splitPane );
	return fullPanel;
    }

    public void windowClosing( WindowEvent event ) {
        exit( );
    }

    public void windowIconified( WindowEvent event ) { }
    public void windowOpened( WindowEvent event ) { }
    public void windowDeiconified( WindowEvent event ) { }
    public void windowDeactivated( WindowEvent event ) { }
    public void windowClosed( WindowEvent event ) { }
    public void windowActivated( WindowEvent event ) { }

    public boolean notCompiled( ) {
	BreakpointFile next = null;
        boolean compiled = true;
	for( Enumeration elts = elements( ); 
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( !next.compiled && next.getFilename( ).endsWith( ".java" ) ) 
		compiled = false;
	}
	return !compiled;
    }

    public boolean textChanged( ) {
	BreakpointFile next = null;
        boolean changed = false;
	for( Enumeration elts = elements( ); 
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.textChanged ) 
		changed = true;
	}
	return changed;
    }

    public void exit( ) {
	BreakpointFile next = null;
        boolean shouldExit = true;
	for( Enumeration elts = elements( ); 
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.textChanged ) {
		int option = DefaultDialog.yesNoCancel( "Save source code in file\n" 
							+ next.getFilename( )+"\n"
							+ "before exiting?" );
		if( option == JOptionPane.YES_OPTION ) {
		    if( !next.save( ) )
			shouldExit = false;
		}
		if( option == JOptionPane.CANCEL_OPTION )
		    return;
	    }
	}
	if( shouldExit ) {
	    if( menuBar.getCurrentMode( ).equals( "Debug" ) )
		messageReceiver.deliverMessage( "quit" );
	    messageReceiver.deliverMessage( "stopexecute" );
	    System.exit( 0 );
	}
    }

    /* take a filename and extract if possible the classname
     * i.e. /home/jjc/bdbj/BdbjGUI.java --> BdbjGUI
     * if not possible return an empty string 
     */

    private String getFile( String filename ) {
	String separator = System.getProperty( "file.separator" );
	String end;
	int lastSep = filename.lastIndexOf( separator );
	if( lastSep == -1 )
	    end = filename;
	else
	    end = filename.substring( lastSep + separator.length( ) );
	return end;
    }


    private void realOpenFile( String filename ) {
	String debugname = getFile( filename );
	BreakpointFile newFile; 
	if( debugname.endsWith( ".class" ) ) {
	    debugname = filename.replace( '$', '.' );
	    newFile = new BytecodeBreakpointFile( debugname, filename );
	}
	else 
	    newFile = new BreakpointFile( debugname, filename );
	newFile.loadFile( filename, true ); // we don't care if it exists or not.


	debugname = newFile.packageDetect( ) + debugname;
	newFile.setDebugname( debugname );
	    
	boolean alreadyLoaded = false;
	for( Enumeration elts = elements( );
	     elts.hasMoreElements( );
	     ) 
	    if( ( (BreakpointFile)elts.nextElement( ) ).getDebugname( ).equals( debugname ) )
		alreadyLoaded = true;
	
	if( !alreadyLoaded ) {
	    for( int i = size( ); i >= 0; i-- ) {
		if( i == 0 ) {
		    insertElementAt( file = newFile, 0 );
		    break;
		}
		else if( elementAt( i - 1 ).toString( ).compareTo( newFile.toString( ) ) < 0 ) {
		    insertElementAt( file = newFile, i );
		    break;
		}
	    }
		
	    // now tell the file what's actually going on at the moment
	    file.setMenuBar( menuBar );
	    file.setMenuMode( menuMode );
	    file.setEditable( editable );
	    file.setMessageReceiver( messageReceiver );
	    file.setFont( new Font( "Monospaced", Font.PLAIN, fontSize ) );
	    
	    SourceStatusPanel ssp = new SourceStatusPanel( file );
	    fileCards.add( ssp, debugname );
	    fileCards.showCard( debugname );
	    fileList.setListData( this );		
	    fileList.setSelectedValue( file, true );	    
	    fileList.validate( );
	} else 
	    DefaultDialog.error( "File with this classname already loaded" );
    }
    
    private void openFile( ) {
        JFileChooser fileDialog = new JFileChooser( );
	fileDialog.setCurrentDirectory( new File( System.getProperty( "user.dir" ) ) );
        int ret = fileDialog.showOpenDialog( null );

        if( ret == JFileChooser.APPROVE_OPTION ) {
	    realOpenFile( fileDialog.getSelectedFile( ).getPath( ) );
	}
    }

    private void openBytecode( ) {
	Object input = DefaultDialog.getInput( "Enter classname\n" +
					       "(Use '/' not '.' and omit \".class\")",
					       "" );	
	if( input != null ) {
	    realOpenFile( input.toString() + ".class" );
	}
    }

    private void closeFile( ) {
	if( size( ) > 0 ) {
	    removeElement( file );
	    fileList.setListData( this );
	    if( size( ) > 0 ) {
		fileCards.showCard( elementAt( 0 ).toString( ) );
	    } else
		fileCards.showCard( "!null" );
	    fileCards.remove( file );
	    if( size( ) > 0 ) {
		fileList.setSelectedIndex( 0 );
		file = (BreakpointFile)elementAt( 0 );
	    } else
		file = null;
	} 
    }

    private String compileCommand = System.getProperty( "bdbj.compile" );
    private String kaffeLocation = System.getProperty( "bdbj.kaffedir" );
    private static String sep = System.getProperty( "file.separator" );

    private String[] vectorToStringArray( Vector vector ) {
	String[] result = new String[ vector.size( ) ];
	for( int i = 0; i < vector.size( ); i++ )
	    result[ i ] = vector.elementAt( i ).toString( );	
	return result;
    }

    private boolean beautify( String filename ) {
	if( kaffeLocation == null )
	    kaffeLocation = ".";
	final Vector argumentsVector = new Vector( );

	argumentsVector.add( kaffeLocation + sep + "bin" + sep + "kaffe" );
	argumentsVector.add( "at.dms.kopi.Main" );
	argumentsVector.add( "-b" );
	argumentsVector.add( filename );

      	try {
	    Process process = Runtime.getRuntime( ).exec( vectorToStringArray( argumentsVector ) );
	    return process.waitFor( ) == 0;
	} catch( IOException exception ) {
	    return false;
	} catch( InterruptedException exception ) {
	    return false;
	}
    }	    

    private boolean compile( String filesToCompile, BreakpointFile file ) {
	// if compiling one file, file is its BreakpointFile object,
	// if compiling all files, file is null.
        Process process;
        try {
	    if( compileCommand == null || compileCommand.trim( ).equals( "" ) )
		compileCommand = "javac -g";
            process = Runtime.getRuntime().exec( compileCommand + " " + filesToCompile );
            BufferedReader inputStream_err = new BufferedReader( new InputStreamReader( process.getErrorStream( ) ) );
            String next;
            StringBuffer errors = new StringBuffer( "" );
            while( ( next = inputStream_err.readLine( ) ) != null )
                errors = errors.append( next + "\n" );

	    // user might be using a compiler other than javac which
	    // sends things to stdout instead of stderr.

	    // if it sends things to both they are separated
	    // and not interleaved - possibly a bad idea.

	    BufferedReader inputStream_out = new BufferedReader( new InputStreamReader( process.getInputStream( ) ) );
	    while( ( next = inputStream_out.readLine( ) ) != null )
                errors = errors.append( next = "\n" );

            if( process.waitFor( ) != 0 ) {
                new JavacErrors( watchTablePane.getBounds( ), errors.toString( ) ); 
		return false;
	    }
            else { 
		if( file != null )
		    file.setCompiled( true );
		else {
		    for( Enumeration elts = elements( ); 
			 elts.hasMoreElements( );
			 ) {
			( (BreakpointFile)elts.nextElement( ) ).setCompiled( true );
		    }
		}
		regenerateAllClassFiles( );
                DefaultDialog.info( "Successful compilation" );
		return true;
	    }
        } catch( IOException e ) {
            DefaultDialog.error( "I/O Exception" );
        } catch( InterruptedException e ) {
            DefaultDialog.error( "Compilation interrupted" );
        }
	return false;
    }
    
    private boolean inCurrentDir( String filename ) {
	File currentDir = new File( System.getProperty( "user.dir" ) );
	File filesParent = new File( new File( filename ).getAbsolutePath( ) ).getParentFile( );
	if( currentDir == null || filesParent == null )
	    return false;
	else
	    return filesParent.equals( currentDir );
    }

    public boolean allSaveCompile( ) {
	BreakpointFile next = null;
        String filesToCompile = "";
	for( Enumeration elts = elements( ); 
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    if( next.getFilename( ).endsWith( ".java" ) ) {
		if( next.save( ) ) {
		    // if( inCurrentDir( next.getFilename( ) ) )
		    filesToCompile += " " + next.getFilename( );
		} else
		    return false;
	    }
	}
	return compile( filesToCompile, null );
    }

    public void regenerateAllClassFiles( ) {
        Object next = null;
	for( Enumeration elts = elements( ); 
	     elts.hasMoreElements( );
	     ) {
	    next = elts.nextElement( );
	    if( next instanceof BytecodeBreakpointFile )		
		((BytecodeBreakpointFile)next).regenerate( );
	}
    }

    private void allSave( ) {
	BreakpointFile next = null;
	for( Enumeration elts = elements( ); 
	     elts.hasMoreElements( );
	     ) 
	    ((BreakpointFile)elts.nextElement( )).save( );
    }

    private void saveCompile( ) {
	if( file != null )
	    if( file.save( ) )
		compile( file.getFilename( ), file );
    }

    private void save( ) {
	if( file != null )
	    file.save( );
    }

    private void markCompiled( ) {
	if( file != null ) {
	    file.setTextChanged( false );
	    file.setCompiled( true );
	}
    }

    private void beautify( ) {
	if( file != null ) {
	    if( beautify( file.getFilename( ) ) ) {
		String newname = new File( file.getFilename( ) ).getName( );
		file.loadFile( newname + ".gen", false );
		new File( newname + ".gen" ).delete( );
		file.setTextChanged( true );
		file.setCompiled( false );
	    }
	}
    }

    private void setFontSize( String size ) {
        int fs = 0;
        try {
            fs = Integer.parseInt( size.trim( ) );
        } catch( NumberFormatException e ) {
            DefaultDialog.error( "Invalid number" );
            return;
        }
        if( fs < 8 || fs > 64 ) {
            DefaultDialog.error( "Invalid font size" );
            return;
        }
	fontSize = fs;
	// now change all the open windows.
	BreakpointFile next = null;
	for( Enumeration elts = elements( ); 
	     elts.hasMoreElements( );
	     ) {
	    next = (BreakpointFile)elts.nextElement( );
	    next.setFont( new Font( "Monospaced", Font.PLAIN, fontSize ) );
	}
	

    }

    public void actionPerformed( ActionEvent event ) {
	String label = event.getActionCommand( );
	if( label.equals( "Open" ) )
	    openFile( );
	else if( label.equals( "Open Bytecode" ) )
	    openBytecode( );
	else if( label.equals( "Close" ) )
	    closeFile( );
	else if( label.equals( "Save and Compile" ) )	   
	    saveCompile( );
	else if( label.equals( "Save" ) )
	    save( );
	else if( label.equals( "Save and Compile All" ) )
	    allSaveCompile( );
	else if( label.equals( "Save All" ) )
	    allSave( );
	else if( label.equals( "Exit" ) )
	    exit( );
	else if( label.equals( "Clear All Breakpoints" ) )
	    clearAllBreakpoints( );
	else if( label.equals( "Mark File as Compiled" ) )
	    markCompiled( );
	else if( label.equals( "Beautify File" ) )
	    beautify( );
	else if( label.equals( "Set Font Size" ) ) {
	    Object input = DefaultDialog.getInput( "Enter font size in points", "" + fontSize );
	    if( input != null )
		setFontSize( input.toString( ) );
	}
	else if( file != null )
	    file.actionPerformed( event );
    }

    public MenuData getFileMenu( ) {
        boolean[] justEdit = { true, false, true };
        boolean[] bothModes = { true, true, true };
        MenuItemData[] controlMenuData =
	    { new MenuItemData( "Open...", "Open", this, "alt O", justEdit ),
	      new MenuItemData( "Open Bytecode...", "Open Bytecode", this, "", justEdit ),
              new MenuItemData( "Save and Compile", "Save and Compile", this, "alt C", justEdit ),
	      new MenuItemData( "Save", "Save", this, "alt shift S", justEdit ),
              new MenuItemData( "Save Backup...", "Save Backup", this, "alt S", justEdit ),
              new MenuItemData( "Restore Backup...", "Restore Backup", this, "alt R", justEdit ),
	      new MenuItemData( "Beautify File", "Beautify File", this, "", justEdit ),
	      new MenuItemData( "Close", "Close", this, "alt Q", justEdit ),
	      new MenuItemData( "", "", null ),
	      new MenuItemData( "Mark File as Compiled", "Mark File as Compiled", this, "", justEdit ),
              new MenuItemData( "", "", null ),	      
	      new MenuItemData( "Save and Compile All", "Save and Compile All", this, "shift alt C", justEdit ),
	      new MenuItemData( "Save All", "Save All", this, "", justEdit ),
	      new MenuItemData( "", "", null ),
              new MenuItemData( "Exit", "Exit", this, "alt X", bothModes ) };
        return new MenuData( "File", controlMenuData );
    }    

    public MenuData getBreakpointMenu( ) {
        boolean[] notTransition = { true, true, false };
        MenuItemData[] breakpointMenuData =
            { new MenuItemData( "Toggle Breakpoint", "Toggle Breakpoint", this, "alt 8", notTransition ),
	      new MenuItemData( "Clear All Breakpoints in File", "Clear All Breakpoints in File", this, "", 
				notTransition ),
              new MenuItemData( "Clear All Breakpoints", "Clear All Breakpoints", this, "alt shift 8", notTransition ) }; 
        return new MenuData( "Breakpoints", breakpointMenuData );    
    }

    public ToolBarData getFileToolBar( ) {
	boolean[] justDebug = { false, true, false };
	boolean[] justEdit = { true, false, false };
	boolean[] bothModes = { true, true, true };
        boolean[] notTransition = { true, true, false };
	ToolBarButtonData[] fileToolBarData =
	    {
		new ToolBarButtonData( "images/open.gif", "Open", this, justEdit ),
		new ToolBarButtonData( "images/save.gif", "Save and Compile All", this, justEdit ),
		new ToolBarButtonData( "", "", null ),
		new ToolBarButtonData( "images/cut.gif", "Cut", this, bothModes ),
		new ToolBarButtonData( "images/copy.gif", "Copy", this, bothModes ),
		new ToolBarButtonData( "images/paste.gif", "Paste", this, bothModes ),
		new ToolBarButtonData( "", "", null ),
		new ToolBarButtonData( "images/toggle.gif", "Toggle Breakpoint", this, notTransition ),
		new ToolBarButtonData( "images/clearallbpts.gif", "Clear All Breakpoints", this, notTransition )
	    };
	return new ToolBarData( "File", fileToolBarData );
    }

    public MenuData getEditMenu( ) {
        boolean[] bothModes = { true, true, true };
        MenuItemData[] editMenuData =
            { new MenuItemData( "Undo", "Undo", this, "control Z", bothModes ),
	      new MenuItemData( "Redo", "Redo", this, "control Y", bothModes ),
	      new MenuItemData( "", "", null ),
	      new MenuItemData( "Cut", "Cut", this, "control X", bothModes ),
	      new MenuItemData( "Copy", "Copy", this, "control C", bothModes ),
	      new MenuItemData( "Paste", "Paste", this, "control V", bothModes ),
	      new MenuItemData( "", "", null ),
	      new MenuItemData( "Goto Line...", "Goto Line", this, "alt G", bothModes ),
              new MenuItemData( "Search...", "Search", this, "alt F", bothModes ), 
              new MenuItemData( "Search Again", "Search Again", this, "alt shift F", bothModes ),
	      new MenuItemData( "", "", null ),
	      new MenuItemData( "Set Font Size", "Set Font Size", this, "", bothModes ) };
        return new MenuData( "Edit", editMenuData );
    }
}

class JavacErrors extends JFrame implements ActionListener {
    private JTextArea textArea = new JTextArea( 20, 80 );
    private Rectangle bounds = null;

    public JavacErrors( Rectangle bounds, String errors ) {
        super( "javac Output" );
        textArea.setFont( new Font( "Monospaced", Font.PLAIN, 14 ) );
	textArea.setText( errors );
	this.bounds = bounds;
        getContentPane( ).setLayout( new BorderLayout( ) );
        getContentPane( ).add( "Center", new JScrollPane( textArea ) );
        
        JButton button = new JButton( "Dismiss" );
        getContentPane( ).add( "South", button );
        button.addActionListener( this );
	pack( );
	setVisible( true );
    }

    public void showText( String errors ) {
        textArea.setText( errors );
	pack( );
        setVisible( true );
    }

    public void actionPerformed( ActionEvent e ) {
        setVisible( false );
	dispose( );
    }
}

class ListPanel extends JPanel {
    public ListPanel( String title, JList list ) {
	super( );
	list.setVisibleRowCount( 10 );
	setLayout( new BorderLayout( ) );
	add( "North", new JLabel( title ) );
	add( "Center", list );
    }
}

class FileCards extends JPanel {
    private CardLayout cardLayout = new CardLayout( );

    public FileCards( ) {
	setLayout( cardLayout );
    }

    public void showCard( String name ) {
	cardLayout.show( this, name );
    }
}
